import math
import numpy as np
import matplotlib
import cv2
import os
from typing import List, Tuple, Union, Optional

from .body import BodyResult, Keypoint

eps = 0.01


def smart_resize(x, s):
    Ht, Wt = s
    if x.ndim == 2:
        Ho, Wo = x.shape
        Co = 1
    else:
        Ho, Wo, Co = x.shape
    if Co == 3 or Co == 1:
        k = float(Ht + Wt) / float(Ho + Wo)
        return cv2.resize(x, (int(Wt), int(Ht)), interpolation=cv2.INTER_AREA if k < 1 else cv2.INTER_LANCZOS4)
    else:
        return np.stack([smart_resize(x[:, :, i], s) for i in range(Co)], axis=2)


def smart_resize_k(x, fx, fy):
    if x.ndim == 2:
        Ho, Wo = x.shape
        Co = 1
    else:
        Ho, Wo, Co = x.shape
    Ht, Wt = Ho * fy, Wo * fx
    if Co == 3 or Co == 1:
        k = float(Ht + Wt) / float(Ho + Wo)
        return cv2.resize(x, (int(Wt), int(Ht)), interpolation=cv2.INTER_AREA if k < 1 else cv2.INTER_LANCZOS4)
    else:
        return np.stack([smart_resize_k(x[:, :, i], fx, fy) for i in range(Co)], axis=2)


def padRightDownCorner(img, stride, padValue):
    h = img.shape[0]
    w = img.shape[1]

    pad = 4 * [None]
    pad[0] = 0 # up
    pad[1] = 0 # left
    pad[2] = 0 if (h % stride == 0) else stride - (h % stride) # down
    pad[3] = 0 if (w % stride == 0) else stride - (w % stride) # right

    img_padded = img
    pad_up = np.tile(img_padded[0:1, :, :]*0 + padValue, (pad[0], 1, 1))
    img_padded = np.concatenate((pad_up, img_padded), axis=0)
    pad_left = np.tile(img_padded[:, 0:1, :]*0 + padValue, (1, pad[1], 1))
    img_padded = np.concatenate((pad_left, img_padded), axis=1)
    pad_down = np.tile(img_padded[-2:-1, :, :]*0 + padValue, (pad[2], 1, 1))
    img_padded = np.concatenate((img_padded, pad_down), axis=0)
    pad_right = np.tile(img_padded[:, -2:-1, :]*0 + padValue, (1, pad[3], 1))
    img_padded = np.concatenate((img_padded, pad_right), axis=1)

    return img_padded, pad


def transfer(model, model_weights):
    transfered_model_weights = {}
    for weights_name in model.state_dict().keys():
        transfered_model_weights[weights_name] = model_weights['.'.join(weights_name.split('.')[1:])]
    return transfered_model_weights


def is_normalized(keypoints: List[Optional[Keypoint]]) -> bool:
    point_normalized = [
        0 <= abs(k.x) <= 1 and 0 <= abs(k.y) <= 1 
        for k in keypoints 
        if k is not None
    ]
    if not point_normalized:
        return False
    return all(point_normalized)

    
def draw_bodypose(canvas: np.ndarray, keypoints: List[Keypoint]) -> np.ndarray:
    """
    Draw keypoints and limbs representing body pose on a given canvas.

    Args:
        canvas (np.ndarray): A 3D numpy array representing the canvas (image) on which to draw the body pose.
        keypoints (List[Keypoint]): A list of Keypoint objects representing the body keypoints to be drawn.

    Returns:
        np.ndarray: A 3D numpy array representing the modified canvas with the drawn body pose.

    Note:
        The function expects the x and y coordinates of the keypoints to be normalized between 0 and 1.
    """
    if not is_normalized(keypoints):
        H, W = 1.0, 1.0
    else:
        H, W, _ = canvas.shape

    stickwidth = 4

    limbSeq = [
        [2, 3], [2, 6], [3, 4], [4, 5], 
        [6, 7], [7, 8], [2, 9], [9, 10], 
        [10, 11], [2, 12], [12, 13], [13, 14], 
        [2, 1], [1, 15], [15, 17], [1, 16], 
        [16, 18],
    ]

    colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0], [255, 255, 0], [170, 255, 0], [85, 255, 0], [0, 255, 0], \
              [0, 255, 85], [0, 255, 170], [0, 255, 255], [0, 170, 255], [0, 85, 255], [0, 0, 255], [85, 0, 255], \
              [170, 0, 255], [255, 0, 255], [255, 0, 170], [255, 0, 85]]

    for (k1_index, k2_index), color in zip(limbSeq, colors):
        keypoint1 = keypoints[k1_index - 1]
        keypoint2 = keypoints[k2_index - 1]

        if keypoint1 is None or keypoint2 is None:
            continue

        Y = np.array([keypoint1.x, keypoint2.x]) * float(W)
        X = np.array([keypoint1.y, keypoint2.y]) * float(H)
        mX = np.mean(X)
        mY = np.mean(Y)
        length = ((X[0] - X[1]) ** 2 + (Y[0] - Y[1]) ** 2) ** 0.5
        angle = math.degrees(math.atan2(X[0] - X[1], Y[0] - Y[1]))
        polygon = cv2.ellipse2Poly((int(mY), int(mX)), (int(length / 2), stickwidth), int(angle), 0, 360, 1)
        cv2.fillConvexPoly(canvas, polygon, [int(float(c) * 0.6) for c in color])

    for keypoint, color in zip(keypoints, colors):
        if keypoint is None:
            continue

        x, y = keypoint.x, keypoint.y
        x = int(x * W)
        y = int(y * H)
        cv2.circle(canvas, (int(x), int(y)), 4, color, thickness=-1)

    return canvas


def draw_handpose(canvas: np.ndarray, keypoints: Union[List[Keypoint], None]) -> np.ndarray:
    """
    Draw keypoints and connections representing hand pose on a given canvas.

    Args:
        canvas (np.ndarray): A 3D numpy array representing the canvas (image) on which to draw the hand pose.
        keypoints (List[Keypoint]| None): A list of Keypoint objects representing the hand keypoints to be drawn
                                          or None if no keypoints are present.

    Returns:
        np.ndarray: A 3D numpy array representing the modified canvas with the drawn hand pose.

    Note:
        The function expects the x and y coordinates of the keypoints to be normalized between 0 and 1.
    """
    if not keypoints:
        return canvas
    
    if not is_normalized(keypoints):
        H, W = 1.0, 1.0
    else:
        H, W, _ = canvas.shape

    edges = [[0, 1], [1, 2], [2, 3], [3, 4], [0, 5], [5, 6], [6, 7], [7, 8], [0, 9], [9, 10], \
             [10, 11], [11, 12], [0, 13], [13, 14], [14, 15], [15, 16], [0, 17], [17, 18], [18, 19], [19, 20]]

    for ie, (e1, e2) in enumerate(edges):
        k1 = keypoints[e1]
        k2 = keypoints[e2]
        if k1 is None or k2 is None:
            continue
        
        x1 = int(k1.x * W)
        y1 = int(k1.y * H)
        x2 = int(k2.x * W)
        y2 = int(k2.y * H)
        if x1 > eps and y1 > eps and x2 > eps and y2 > eps:
            cv2.line(canvas, (x1, y1), (x2, y2), matplotlib.colors.hsv_to_rgb([ie / float(len(edges)), 1.0, 1.0]) * 255, thickness=2)

    for keypoint in keypoints:
        if keypoint is None:
            continue

        x, y = keypoint.x, keypoint.y
        x = int(x * W)
        y = int(y * H)
        if x > eps and y > eps:
            cv2.circle(canvas, (x, y), 4, (0, 0, 255), thickness=-1)
    return canvas


def draw_facepose(canvas: np.ndarray, keypoints: Union[List[Keypoint], None]) -> np.ndarray:
    """
    Draw keypoints representing face pose on a given canvas.

    Args:
        canvas (np.ndarray): A 3D numpy array representing the canvas (image) on which to draw the face pose.
        keypoints (List[Keypoint]| None): A list of Keypoint objects representing the face keypoints to be drawn
                                          or None if no keypoints are present.

    Returns:
        np.ndarray: A 3D numpy array representing the modified canvas with the drawn face pose.

    Note:
        The function expects the x and y coordinates of the keypoints to be normalized between 0 and 1.
    """    
    if not keypoints:
        return canvas
    
    if not is_normalized(keypoints):
        H, W = 1.0, 1.0
    else:
        H, W, _ = canvas.shape

    for keypoint in keypoints:
        if keypoint is None:
            continue
        
        x, y = keypoint.x, keypoint.y
        x = int(x * W)
        y = int(y * H)
        if x > eps and y > eps:
            cv2.circle(canvas, (x, y), 3, (255, 255, 255), thickness=-1)
    return canvas


# detect hand according to body pose keypoints
# please refer to https://github.com/CMU-Perceptual-Computing-Lab/openpose/blob/master/src/openpose/hand/handDetector.cpp
def handDetect(body: BodyResult, oriImg) -> List[Tuple[int, int, int, bool]]:
    """
    Detect hands in the input body pose keypoints and calculate the bounding box for each hand.

    Args:
        body (BodyResult): A BodyResult object containing the detected body pose keypoints.
        oriImg (numpy.ndarray): A 3D numpy array representing the original input image.

    Returns:
        List[Tuple[int, int, int, bool]]: A list of tuples, each containing the coordinates (x, y) of the top-left
                                          corner of the bounding box, the width (height) of the bounding box, and
                                          a boolean flag indicating whether the hand is a left hand (True) or a
                                          right hand (False).

    Notes:
        - The width and height of the bounding boxes are equal since the network requires squared input.
        - The minimum bounding box size is 20 pixels.
    """
    ratioWristElbow = 0.33
    detect_result = []
    image_height, image_width = oriImg.shape[0:2]
    
    keypoints = body.keypoints
    # right hand: wrist 4, elbow 3, shoulder 2
    # left hand: wrist 7, elbow 6, shoulder 5
    left_shoulder = keypoints[5]
    left_elbow = keypoints[6]
    left_wrist = keypoints[7]
    right_shoulder = keypoints[2]
    right_elbow = keypoints[3]
    right_wrist = keypoints[4]

    # if any of three not detected
    has_left = all(keypoint is not None for keypoint in (left_shoulder, left_elbow, left_wrist))
    has_right = all(keypoint is not None for keypoint in (right_shoulder, right_elbow, right_wrist))
    if not (has_left or has_right):
        return []
    
    hands = []
    #left hand
    if has_left:
        hands.append([
            left_shoulder.x, left_shoulder.y,
            left_elbow.x, left_elbow.y,
            left_wrist.x, left_wrist.y,
            True
        ])
    # right hand
    if has_right:
        hands.append([
            right_shoulder.x, right_shoulder.y,
            right_elbow.x, right_elbow.y,
            right_wrist.x, right_wrist.y,
            False
        ])

    for x1, y1, x2, y2, x3, y3, is_left in hands:
        # pos_hand = pos_wrist + ratio * (pos_wrist - pos_elbox) = (1 + ratio) * pos_wrist - ratio * pos_elbox
        # handRectangle.x = posePtr[wrist*3] + ratioWristElbow * (posePtr[wrist*3] - posePtr[elbow*3]);
        # handRectangle.y = posePtr[wrist*3+1] + ratioWristElbow * (posePtr[wrist*3+1] - posePtr[elbow*3+1]);
        # const auto distanceWristElbow = getDistance(poseKeypoints, person, wrist, elbow);
        # const auto distanceElbowShoulder = getDistance(poseKeypoints, person, elbow, shoulder);
        # handRectangle.width = 1.5f * fastMax(distanceWristElbow, 0.9f * distanceElbowShoulder);
        x = x3 + ratioWristElbow * (x3 - x2)
        y = y3 + ratioWristElbow * (y3 - y2)
        distanceWristElbow = math.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2)
        distanceElbowShoulder = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        width = 1.5 * max(distanceWristElbow, 0.9 * distanceElbowShoulder)
        # x-y refers to the center --> offset to topLeft point
        # handRectangle.x -= handRectangle.width / 2.f;
        # handRectangle.y -= handRectangle.height / 2.f;
        x -= width / 2
        y -= width / 2  # width = height
        # overflow the image
        if x < 0: x = 0
        if y < 0: y = 0
        width1 = width
        width2 = width
        if x + width > image_width: width1 = image_width - x
        if y + width > image_height: width2 = image_height - y
        width = min(width1, width2)
        # the max hand box value is 20 pixels
        if width >= 20:
            detect_result.append((int(x), int(y), int(width), is_left))

    '''
    return value: [[x, y, w, True if left hand else False]].
    width=height since the network require squared input.
    x, y is the coordinate of top left 
    '''
    return detect_result


# Written by Lvmin
def faceDetect(body: BodyResult, oriImg) -> Union[Tuple[int, int, int], None]:
    """
    Detect the face in the input body pose keypoints and calculate the bounding box for the face.

    Args:
        body (BodyResult): A BodyResult object containing the detected body pose keypoints.
        oriImg (numpy.ndarray): A 3D numpy array representing the original input image.

    Returns:
        Tuple[int, int, int] | None: A tuple containing the coordinates (x, y) of the top-left corner of the
                                   bounding box and the width (height) of the bounding box, or None if the
                                   face is not detected or the bounding box width is less than 20 pixels.

    Notes:
        - The width and height of the bounding box are equal.
        - The minimum bounding box size is 20 pixels.
    """
    # left right eye ear 14 15 16 17
    image_height, image_width = oriImg.shape[0:2]
    
    keypoints = body.keypoints
    head = keypoints[0]
    left_eye = keypoints[14]
    right_eye = keypoints[15]
    left_ear = keypoints[16]
    right_ear = keypoints[17]
    
    if head is None or all(keypoint is None for keypoint in (left_eye, right_eye, left_ear, right_ear)):
        return None

    width = 0.0
    x0, y0 = head.x, head.y

    if left_eye is not None:
        x1, y1 = left_eye.x, left_eye.y
        d = max(abs(x0 - x1), abs(y0 - y1))
        width = max(width, d * 3.0)

    if right_eye is not None:
        x1, y1 = right_eye.x, right_eye.y
        d = max(abs(x0 - x1), abs(y0 - y1))
        width = max(width, d * 3.0)

    if left_ear is not None:
        x1, y1 = left_ear.x, left_ear.y
        d = max(abs(x0 - x1), abs(y0 - y1))
        width = max(width, d * 1.5)

    if right_ear is not None:
        x1, y1 = right_ear.x, right_ear.y
        d = max(abs(x0 - x1), abs(y0 - y1))
        width = max(width, d * 1.5)

    x, y = x0, y0

    x -= width
    y -= width

    if x < 0:
        x = 0

    if y < 0:
        y = 0

    width1 = width * 2
    width2 = width * 2

    if x + width > image_width:
        width1 = image_width - x

    if y + width > image_height:
        width2 = image_height - y

    width = min(width1, width2)

    if width >= 20:
        return int(x), int(y), int(width)
    else:
        return None


# get max index of 2d array
def npmax(array):
    arrayindex = array.argmax(1)
    arrayvalue = array.max(1)
    i = arrayvalue.argmax()
    j = arrayindex[i]
    return i, j

def guess_onnx_input_shape_dtype(filename):
    dtype = np.float32
    if "fp16" in filename:
        dtype = np.float16
    elif "int8" in filename:
        dtype = np.uint8
    input_size = (640, 640) if "yolo" in filename else (192, 256)
    if "384" in filename:
        input_size = (288, 384)
    elif "256" in filename:
        input_size = (256, 256)
    return input_size, dtype

if os.getenv('AUX_ORT_PROVIDERS'):
    ONNX_PROVIDERS = os.getenv('AUX_ORT_PROVIDERS').split(',')
else:
    ONNX_PROVIDERS = ["CUDAExecutionProvider", "DirectMLExecutionProvider", "OpenVINOExecutionProvider", "ROCMExecutionProvider", "CPUExecutionProvider"]
def get_ort_providers() -> List[str]:
    providers = []
    try:
        import onnxruntime as ort
        for provider in ONNX_PROVIDERS:
            if provider in ort.get_available_providers():
                providers.append(provider)
        return providers
    except:
        return []

def is_model_torchscript(model) -> bool:
    return bool(type(model).__name__ == "RecursiveScriptModule")

def get_model_type(Nodesname, filename) -> str:
    ort_providers = list(filter(lambda x : x != "CPUExecutionProvider", get_ort_providers()))
    if filename is None:
        return None
    elif ("onnx" in filename) and ort_providers:
        print(f"{Nodesname}: Caching ONNXRuntime session {filename}...")
        return "ort"
    elif ("onnx" in filename):
        print(f"{Nodesname}: Caching OpenCV DNN module {filename} on cv2.DNN...")
        return "cv2"
    else:
        print(f"{Nodesname}: Caching TorchScript module {filename} on ...")
        return  "torchscript"



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum




class Llama3_8B(ReplicateNode):
    """Base version of Llama 3, an 8 billion parameter language model from Meta."""

    @classmethod
    def replicate_model_id(cls): return 'meta/meta-llama-3-8b:9a9e68fc8695f5847ce944a5cecf9967fd7c64d0fb8c8af1d5bdcc71f03c5e47'
    @classmethod
    def get_hardware(cls): return 'None'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/dd9ac11a-edda-4d33-b413-6a721c44dfb0/meta-logo.png', 'created_at': '2024-04-17T18:04:26.049832Z', 'description': 'Base version of Llama 3, an 8 billion parameter language model from Meta.', 'github_url': 'https://github.com/meta-llama/llama3', 'license_url': 'https://github.com/meta-llama/llama3/blob/main/LICENSE', 'name': 'meta-llama-3-8b', 'owner': 'meta', 'paper_url': None, 'run_count': 36298455, 'url': 'https://replicate.com/meta/meta-llama-3-8b', 'visibility': 'public'}
    @classmethod
    def return_type(cls): return str


    top_k: int = Field(title='Top K', description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).', default=50)
    top_p: float = Field(title='Top P', description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).', default=0.9)
    prompt: str = Field(title='Prompt', description='Prompt', default='')
    max_tokens: int = Field(title='Max Tokens', description='The maximum number of tokens the model should generate as output.', default=512)
    min_tokens: int = Field(title='Min Tokens', description='The minimum number of tokens the model should generate as output.', default=0)
    temperature: float = Field(title='Temperature', description='The value used to modulate the next token probabilities.', default=0.6)
    prompt_template: str = Field(title='Prompt Template', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.', default='{prompt}')
    presence_penalty: float = Field(title='Presence Penalty', description='Presence penalty', default=1.15)
    frequency_penalty: float = Field(title='Frequency Penalty', description='Frequency penalty', default=0.2)



class Llama3_70B(ReplicateNode):
    """Base version of Llama 3, a 70 billion parameter language model from Meta."""

    @classmethod
    def replicate_model_id(cls): return 'meta/meta-llama-3-70b:83c5bdea9941e83be68480bd06ad792f3f295612a24e4678baed34083083a87f'
    @classmethod
    def get_hardware(cls): return 'None'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/8e044b4c-0b20-4717-83bd-a94d89fb0dbe/meta-logo.png', 'created_at': '2024-04-17T18:05:18.044746Z', 'description': 'Base version of Llama 3, a 70 billion parameter language model from Meta.', 'github_url': None, 'license_url': 'https://github.com/meta-llama/llama3/blob/main/LICENSE', 'name': 'meta-llama-3-70b', 'owner': 'meta', 'paper_url': None, 'run_count': 278560, 'url': 'https://replicate.com/meta/meta-llama-3-70b', 'visibility': 'public'}
    @classmethod
    def return_type(cls): return str


    top_k: int = Field(title='Top K', description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).', default=50)
    top_p: float = Field(title='Top P', description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).', default=0.9)
    prompt: str = Field(title='Prompt', description='Prompt', default='')
    max_tokens: int = Field(title='Max Tokens', description='The maximum number of tokens the model should generate as output.', default=512)
    min_tokens: int = Field(title='Min Tokens', description='The minimum number of tokens the model should generate as output.', default=0)
    temperature: float = Field(title='Temperature', description='The value used to modulate the next token probabilities.', default=0.6)
    prompt_template: str = Field(title='Prompt Template', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.', default='{prompt}')
    presence_penalty: float = Field(title='Presence Penalty', description='Presence penalty', default=1.15)
    frequency_penalty: float = Field(title='Frequency Penalty', description='Frequency penalty', default=0.2)



class Llama3_8B_Instruct(ReplicateNode):
    """An 8 billion parameter language model from Meta, fine tuned for chat completions"""

    @classmethod
    def replicate_model_id(cls): return 'meta/meta-llama-3-8b-instruct:5a6809ca6288247d06daf6365557e5e429063f32a21146b2a807c682652136b8'
    @classmethod
    def get_hardware(cls): return 'None'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/927d3fce-75e3-4af9-92da-f537bc34072a/meta-logo.png', 'created_at': '2024-04-17T21:44:58.480057Z', 'description': 'An 8 billion parameter language model from Meta, fine tuned for chat completions', 'github_url': 'https://github.com/meta-llama/llama3', 'license_url': 'https://github.com/meta-llama/llama3/blob/main/LICENSE', 'name': 'meta-llama-3-8b-instruct', 'owner': 'meta', 'paper_url': None, 'run_count': 8555682, 'url': 'https://replicate.com/meta/meta-llama-3-8b-instruct', 'visibility': 'public'}
    @classmethod
    def return_type(cls): return str


    top_k: int = Field(title='Top K', description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).', default=50)
    top_p: float = Field(title='Top P', description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).', default=0.9)
    prompt: str = Field(title='Prompt', description='Prompt', default='')
    max_tokens: int = Field(title='Max Tokens', description='The maximum number of tokens the model should generate as output.', default=512)
    min_tokens: int = Field(title='Min Tokens', description='The minimum number of tokens the model should generate as output.', default=0)
    temperature: float = Field(title='Temperature', description='The value used to modulate the next token probabilities.', default=0.6)
    prompt_template: str = Field(title='Prompt Template', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.', default='{prompt}')
    presence_penalty: float = Field(title='Presence Penalty', description='Presence penalty', default=1.15)
    frequency_penalty: float = Field(title='Frequency Penalty', description='Frequency penalty', default=0.2)



class Llama3_70B_Instruct(ReplicateNode):
    """A 70 billion parameter language model from Meta, fine tuned for chat completions"""

    @classmethod
    def replicate_model_id(cls): return 'meta/meta-llama-3-70b-instruct:fbfb20b472b2f3bdd101412a9f70a0ed4fc0ced78a77ff00970ee7a2383c575d'
    @classmethod
    def get_hardware(cls): return 'None'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/12ed0c29-9236-4a21-ac5a-7faff3045ab1/meta-logo.png', 'created_at': '2024-04-17T21:44:13.482460Z', 'description': 'A 70 billion parameter language model from Meta, fine tuned for chat completions', 'github_url': 'https://github.com/meta-llama/llama3', 'license_url': 'https://github.com/meta-llama/llama3/blob/main/LICENSE', 'name': 'meta-llama-3-70b-instruct', 'owner': 'meta', 'paper_url': None, 'run_count': 28162933, 'url': 'https://replicate.com/meta/meta-llama-3-70b-instruct', 'visibility': 'public'}
    @classmethod
    def return_type(cls): return str


    top_k: int = Field(title='Top K', description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering).', default=50)
    top_p: float = Field(title='Top P', description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751).', default=0.9)
    prompt: str = Field(title='Prompt', description='Prompt', default='')
    max_tokens: int = Field(title='Max Tokens', description='The maximum number of tokens the model should generate as output.', default=512)
    min_tokens: int = Field(title='Min Tokens', description='The minimum number of tokens the model should generate as output.', default=0)
    temperature: float = Field(title='Temperature', description='The value used to modulate the next token probabilities.', default=0.6)
    prompt_template: str = Field(title='Prompt Template', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.', default='{prompt}')
    presence_penalty: float = Field(title='Presence Penalty', description='Presence penalty', default=1.15)
    frequency_penalty: float = Field(title='Frequency Penalty', description='Frequency penalty', default=0.2)



class Snowflake_Arctic_Instruct(ReplicateNode):
    """An efficient, intelligent, and truly open-source language model"""

    @classmethod
    def replicate_model_id(cls): return 'snowflake/snowflake-arctic-instruct:6133ce49e79282dc4fdbd984779921aa2160fb557ab17e0854aa2a7d690afb9c'
    @classmethod
    def get_hardware(cls): return 'None'
    @classmethod
    def model_info(cls): return {'cover_image_url': 'https://tjzk.replicate.delivery/models_models_cover_image/18259dd0-6c42-441b-90f2-b67af9e4320b/Snowflake_Arctic_Opengraph_1200x6.png', 'created_at': '2024-04-24T00:08:29.300675Z', 'description': 'An efficient, intelligent, and truly open-source language model', 'github_url': 'https://github.com/Snowflake-Labs/snowflake-arctic', 'license_url': 'https://www.apache.org/licenses/LICENSE-2.0', 'name': 'snowflake-arctic-instruct', 'owner': 'snowflake', 'paper_url': None, 'run_count': 519892, 'url': 'https://replicate.com/snowflake/snowflake-arctic-instruct', 'visibility': 'public'}
    @classmethod
    def return_type(cls): return str


    top_k: int = Field(title='Top K', description='The number of highest probability tokens to consider for generating the output. If > 0, only keep the top k tokens with highest probability (top-k filtering). Lower to ignore less likely tokens', default=50)
    top_p: float = Field(title='Top P', description='A probability threshold for generating the output. If < 1.0, only keep the top tokens with cumulative probability >= top_p (nucleus filtering). Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751). Lower to ignore less likely tokens.', default=0.9)
    prompt: str = Field(title='Prompt', description='Prompt to send to the model.', default='')
    temperature: float = Field(title='Temperature', description='The value used to modulate the next token probabilities. Adjusts randomness of outputs, greater than 1 is random and 0 is deterministic, 0.75 is a good starting value.', default=0.6)
    max_new_tokens: int = Field(title='Max New Tokens', description='The maximum number of tokens the model should generate as output. A word is generally 2-3 tokens.', default=512)
    min_new_tokens: int = Field(title='Min New Tokens', description='The minimum number of tokens the model should generate as output. A word is generally 2-3 tokens.', default=0)
    stop_sequences: str = Field(title='Stop Sequences', description="A comma-separated list of sequences to stop generation at. For example, '<end>,<stop>' will stop generation at the first instance of 'end' or '<stop>'.", default='<|im_end|>')
    prompt_template: str = Field(title='Prompt Template', description='Prompt template. The string `{prompt}` will be substituted for the input prompt. If you want to generate dialog output, use this template as a starting point and construct the prompt string manually, leaving `prompt_template={prompt}`.', default="<|im_start|>system\nYou're a helpful assistant<|im_end|>\n<|im_start|>user\n{prompt}<|im_end|>\n\n<|im_start|>assistant\n")
    presence_penalty: float = Field(title='Presence Penalty', description='A parameter that penalizes repeated tokens regardless of the number of appearances. As the value increases, the model will be less likely to repeat tokens in the output.', default=1.15)
    frequency_penalty: float = Field(title='Frequency Penalty', description='Frequency penalty is similar to presence penalty, but while presence penalty applies to all tokens that have been sampled at least once, the frequency penalty proportional to how often a particular token has already been sampled.', default=0.2)


from pydantic import Field
from nodetool.nodes.comfy.controlnet import PreprocessImage


class TilePreprocessor(PreprocessImage):
    pyrUp_iters: int = Field(
        default=3,
        description="The number of times to apply pyrUp.",
        ge=1,
        le=10,
    )


class SAMPreprocessor(PreprocessImage):
    pass



import asyncio
from contextlib import contextmanager
from datetime import datetime
import gc

from pydantic import BaseModel

from nodetool.workflows.base_node import GroupNode
from nodetool.workflows.types import NodeProgress, NodeUpdate, WorkflowUpdate
from nodetool.workflows.run_job_request import RunJobRequest

from nodetool.workflows.processing_context import (
    ProcessingContext,
)
from nodetool.workflows.base_node import (
    BaseNode,
)
from nodetool.workflows.base_node import requires_capabilities_from_request
from nodetool.common.environment import Environment
from nodetool.workflows.graph import Graph, topological_sort


log = Environment.get_logger()


class WorkflowRunner:
    """
    The WorkflowRunner is responsible for executing a graph of nodes in a topological order.
    Nodes will be executed asynchronously in parallel, and the results will be collected and returned.
    If the workflow requires gpu, it will be executed on a worker.
    """

    status: str = "running"
    current_node: str | None = None

    def is_running(self):
        return self.status == "running"

    async def run(self, req: RunJobRequest, context: ProcessingContext):
        """
        Evaluates graph nodes in topological order.

        Each level of the graph is processed in parallel.

        Output nodes are collected and returned as a dictionary.
        """
        assert req.graph is not None, "Graph is required"

        graph_capabilities = requires_capabilities_from_request(req)

        graph = Graph.from_dict(
            {
                "nodes": [node.model_dump() for node in req.graph.nodes],
                "edges": [edge.model_dump() for edge in req.graph.edges],
            }
        )
        graph = graph.build_sub_graphs()

        input_nodes = {node.name: node for node in graph.inputs()}
        output_nodes = graph.outputs()

        if req.params:
            for key, value in req.params.items():
                if key not in input_nodes:
                    raise ValueError("No input node found for param: " + str(key))

                node = input_nodes[key]
                node.assign_property("value", value)

        with self.torch_capabilities(graph_capabilities, context):
            await self.process_graph(context, graph)

        output = {}
        for node in output_nodes:
            output[node.name] = context.get_result(node._id, "output")

        context.post_message(WorkflowUpdate(result=output))

        self.status = "completed"

    async def process_graph(self, context: ProcessingContext, graph: Graph):
        """
        Process the graph in a topological order, executing each node in parallel.

        Args:
            context (ProcessingContext): The processing context containing the graph, edges, and nodes.
            graph (Graph): The graph to be processed.

        Returns:
            None
        """
        context.graph = graph
        sorted_nodes = topological_sort(graph.edges, graph.nodes)

        for level in sorted_nodes:
            nodes = [graph.find_node(i) for i in level]
            await asyncio.gather(
                *[self.process_node(context, node) for node in nodes if node]
            )

    async def process_node(self, context: ProcessingContext, node: BaseNode):
        """
        Process a node in the workflow.
        Skip nodes that have already been processed.
        Checks for special types of nodes, such as loop nodes.

        Args:
            context (ProcessingContext): The processing context.
            node (Node): The node to be processed.

        Returns:
            None
        """

        self.current_node = node._id

        # Skip nodes that have already been processed in a sub graph.
        if node._id in context.processed_nodes:
            return

        # If the node is a loop node, run the loop node, which will process the subgraph.
        if isinstance(node, GroupNode):
            await self.run_group_node(node, context)
        else:
            await self.process_regular_node(context, node)

    async def run_group_node(self, group_node: GroupNode, context: ProcessingContext):
        """
        Find all nodes from the subgraph of the loop node and pass it to
        the loop node for processing.
        """
        # Assign the input values from edges to the node properties.
        for name, value in context.get_node_inputs(group_node._id).items():
            group_node.assign_property(name, value)

        await group_node.process_subgraph(
            context=context,
            runner=self,
        )
        context.processed_nodes.add(group_node._id)

    async def process_regular_node(self, context: ProcessingContext, node: BaseNode):
        """
        Processes a single node in the graph.

        This will get all the results from the input nodes and assign
        them to the properties of the node. Then the node is processed
        and the result is stored in the context's results dictionary.

        Once the node is processed, it is added to the completed_tasks set.
        """
        started_at = datetime.now()

        try:
            # Assign the input values from edges to the node properties.
            for name, value in context.get_node_inputs(node.id).items():
                node.assign_property(name, value)

            context.post_message(
                NodeUpdate(
                    node_id=node.id,
                    node_name=node.get_title(),
                    status="running",
                    started_at=started_at.isoformat(),
                )
            )

            result = await node.process(context)
            result = await node.convert_output(context, result)
        except Exception as e:
            context.post_message(
                NodeUpdate(
                    node_id=node.id,
                    node_name=node.get_title(),
                    status="error",
                    error=str(e)[:1000],
                    started_at=started_at.isoformat(),
                    completed_at=datetime.now().isoformat(),
                )
            )
            raise e
        else:
            res_for_update = result.copy()

            for o in node.outputs():
                # Comfy types are not sent to the client because they are not JSON serializable.
                if o.type.is_comfy_type():
                    del res_for_update[o.name]
                if isinstance(res_for_update.get(o.name), BaseModel):
                    res_for_update[o.name] = res_for_update[o.name].model_dump()

            context.post_message(
                NodeUpdate(
                    node_id=node._id,
                    node_name=node.get_title(),
                    status="completed",
                    result=res_for_update,
                    started_at=started_at.isoformat(),
                    completed_at=datetime.now().isoformat(),
                )
            )

        context.processed_nodes.add(node._id)
        context.set_result(node._id, result)

    @contextmanager
    def torch_capabilities(self, capabilities: list[str], context: ProcessingContext):
        if "comfy" in capabilities:
            if not "comfy" in context.capabilities:
                raise ValueError("Comfy nodes are not supported in this environment")

            import comfy.model_management
            import comfy.utils
            import torch

            def hook(value, total, preview_image):
                comfy.model_management.throw_exception_if_processing_interrupted()
                context.post_message(
                    NodeProgress(
                        node_id=self.current_node or "",
                        progress=value,
                        total=total,
                        # preview_image=self.encode_image(preview_image),
                    )
                )

            comfy.utils.set_progress_bar_global_hook(hook)

            # comfy.model_management.unload_all_models()
            log.info(f"VRAM: {torch.cuda.memory_allocated(0) / 1024 / 1024 / 1024}")
            with torch.inference_mode():
                comfy.model_management.cleanup_models()
                yield
                # gc.collect()
                # comfy.model_management.soft_empty_cache()
                log.info(f"VRAM: {torch.cuda.memory_allocated(0) / 1024 / 1024 / 1024}")

        else:
            yield



from pydantic import Field
from nodetool.common.huggingface_node import HuggingfaceNode
from nodetool.workflows.processing_context import ProcessingContext
from enum import Enum


class ModelId(str, Enum):
    FALCONSAI_TEXT_SUMMARIZATION = "Falconsai/text_summarization"
    FALCONSAI_MEDICAL_SUMMARIZATION = "Falconsai/medical_summarization"
    IMVLADIKON_HET5_SUMMARIZATION = "imvladikon/het5_summarization"


class Summarize(HuggingfaceNode):
    """
    Summarization is the task of producing a shorter version of a document while preserving its important information. Some models can extract text from the original input, while other models can generate entirely new text.
    text, summarization, huggingface

    ### Use Cases
    Research papers can be summarized to allow researchers to spend less time selecting which articles to read. There are several approaches you can take for a task like this:

    Use an existing extractive summarization model on the Hub to do inference.
    Pick an existing language model trained for academic papers. This model can then be trained in a process called fine-tuning so it can solve the summarization task.
    Use a sequence-to-sequence model like T5 for abstractive text summarization.
    """

    model: ModelId = Field(
        default=ModelId.FALCONSAI_TEXT_SUMMARIZATION,
        title="Model ID on Huggingface",
        description="The model ID to use for the summarization",
    )
    inputs: str = Field(
        default="",
        title="Inputs",
        description="The input text to the model",
    )

    async def process(self, context: ProcessingContext) -> str:
        result = await self.run_huggingface(
            model_id=self.model.value, context=context, params={"inputs": self.inputs}
        )
        return result[0]["summary_text"]  # type: ignore



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class CLIPLoader(GraphNode):
    clip_name: CLIPFile | GraphNode | tuple[GraphNode, str] = Field(default=CLIPFile(type='comfy.clip_file', name=''), description='The name of the CLIP to load.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.CLIPLoader"



class CLIPSetLastLayer(GraphNode):
    clip: CLIP | GraphNode | tuple[GraphNode, str] = Field(default=CLIP(type='comfy.clip'), description='The CLIP model to modify.')
    stop_at_clip_layer: int | GraphNode | tuple[GraphNode, str] = Field(default=-1, description='The index of the last CLIP layer to use.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.CLIPSetLastLayer"



class CLIPTextEncode(GraphNode):
    text: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The prompt to use.')
    clip: CLIP | GraphNode | tuple[GraphNode, str] = Field(default=CLIP(type='comfy.clip'), description='The CLIP model to use.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.CLIPTextEncode"



class CLIPVisionEncode(GraphNode):
    clip_vision: CLIPVision | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVision(type='comfy.clip_vision'), description='The CLIP vision model to use for encoding.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to encode with the CLIP vision model.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.CLIPVisionEncode"



class ConditioningAverage(GraphNode):
    conditioning_to: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The target conditioning.')
    conditioning_from: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The source conditioning.')
    conditioning_to_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the target conditioning.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningAverage"



class ConditioningCombine(GraphNode):
    conditioning_1: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The first conditioning input.')
    conditioning_2: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The second conditioning input.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningCombine"



class ConditioningConcat(GraphNode):
    conditioning_to: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to concatenate to.')
    conditioning_from: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to concatenate from.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningConcat"



class ConditioningSetArea(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to modify.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=64, description='The width of the area.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=64, description='The height of the area.')
    x: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The x-coordinate of the top-left corner of the area.')
    y: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The y-coordinate of the top-left corner of the area.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the conditioning in the set area.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningSetArea"



class ConditioningSetAreaPercentage(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to modify.')
    width: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The width of the area as a percentage of the total width.')
    height: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The height of the area as a percentage of the total height.')
    x: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The x-coordinate of the top-left corner of the area as a percentage.')
    y: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The y-coordinate of the top-left corner of the area as a percentage.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the conditioning in the set area.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningSetAreaPercentage"


from nodetool.nodes.comfy.conditioning import SetConditioningAreaEnum

class ConditioningSetMask(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to modify.')
    mask: Mask | GraphNode | tuple[GraphNode, str] = Field(default=Mask(type='comfy.mask'), description='The mask to use for setting the conditioning.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the conditioning within the mask.')
    set_cond_area: SetConditioningAreaEnum | GraphNode | tuple[GraphNode, str] = Field(default=SetConditioningAreaEnum('default'), description='Method to determine the area for setting conditioning.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningSetMask"



class ConditioningSetTimestepRange(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to set timestep range.')
    start: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The start of the timestep range.')
    end: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The end of the timestep range.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningSetTimestepRange"



class ConditioningZeroOut(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to be zeroed out.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ConditioningZeroOut"



class ControlNetApply(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to apply.')
    control_net: ControlNet | GraphNode | tuple[GraphNode, str] = Field(default=ControlNet(type='comfy.control_net'), description='The control net to apply.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default='', description='The image to apply to.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the controlnet.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ControlNetApply"



class ControlNetApplyAdvanced(GraphNode):
    positive: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The positive conditioning to apply.')
    negative: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The negative conditioning to apply.')
    control_net: ControlNet | GraphNode | tuple[GraphNode, str] = Field(default=ControlNet(type='comfy.control_net'), description='The ControlNet to use.')
    image: ImageTensor | GraphNode | tuple[GraphNode, str] = Field(default=ImageTensor(type='comfy.image_tensor'), description='The image to apply conditioning adjustments to.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of conditioning.')
    start_percent: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The start percentage from which to apply conditioning.')
    end_percent: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The end percentage until which to apply conditioning.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.ControlNetApplyAdvanced"



class DualCLIPLoader(GraphNode):
    clip_name1: CLIPFile | GraphNode | tuple[GraphNode, str] = Field(default=CLIPFile(type='comfy.clip_file', name=''), description='The name of the CLIP to load.')
    clip_name2: CLIPFile | GraphNode | tuple[GraphNode, str] = Field(default=CLIPFile(type='comfy.clip_file', name=''), description='The name of the CLIP to load.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.DualCLIPLoader"



class GLIGENTextBoxApply(GraphNode):
    conditioning_to: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The input conditioning to modify.')
    clip: CLIP | GraphNode | tuple[GraphNode, str] = Field(default=CLIP(type='comfy.clip'), description='The CLIP instance to use.')
    gligen_textbox_model: GLIGEN | GraphNode | tuple[GraphNode, str] = Field(default=GLIGEN(type='comfy.gligen'), description='The GLIGEN textbox model to apply.')
    text: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The text to apply.')
    width: int | GraphNode | tuple[GraphNode, str] = Field(default=64, description='The width of the text box.')
    height: int | GraphNode | tuple[GraphNode, str] = Field(default=64, description='The height of the text box.')
    x: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The x position of the text box.')
    y: int | GraphNode | tuple[GraphNode, str] = Field(default=0, description='The y position of the text box.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.GLIGENTextBoxApply"



class LoraLoader(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to apply Lora to.')
    clip: CLIP | GraphNode | tuple[GraphNode, str] = Field(default=CLIP(type='comfy.clip'), description='The CLIP model to apply Lora to.')
    lora_name: LORAFile | GraphNode | tuple[GraphNode, str] = Field(default=LORAFile(type='comfy.lora_file', name=''), description='The name of the LoRA to load.')
    strength_model: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the LoRA to apply to the model.')
    strength_clip: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the LoRA to apply to the CLIP.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.LoraLoader"



class LoraLoaderModelOnly(GraphNode):
    model: UNet | GraphNode | tuple[GraphNode, str] = Field(default=UNet(type='comfy.unet'), description='The model to apply Lora to.')
    lora_name: LORAFile | GraphNode | tuple[GraphNode, str] = Field(default=LORAFile(type='comfy.lora_file', name=''), description='The name of the LoRA to load.')
    strength_model: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the LoRA to apply to the model.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.LoraLoaderModelOnly"



class VAELoader(GraphNode):
    vae_name: VAEFile | GraphNode | tuple[GraphNode, str] = Field(default=VAEFile(type='comfy.vae_file', name=''), description='The name of the VAE to load.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.VAELoader"



class unCLIPConditioning(GraphNode):
    conditioning: Conditioning | GraphNode | tuple[GraphNode, str] = Field(default=Conditioning(type='comfy.conditioning'), description='The conditioning to modify.')
    clip_vision_output: CLIPVisionOutput | GraphNode | tuple[GraphNode, str] = Field(default=CLIPVisionOutput(type='comfy.clip_vision_output'), description='The CLIP vision output to associate.')
    strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1.0, description='The strength of the association with the CLIP vision output.')
    noise_augmentation: float | GraphNode | tuple[GraphNode, str] = Field(default=0.0, description='The amount of noise augmentation to apply.')
    @classmethod
    def get_node_type(cls): return "comfy.conditioning.unCLIPConditioning"





from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode


class Concat(GraphNode):
    a: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    b: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.Concat"



class JSONToDataframe(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.JSONToDataframe"



class Join(GraphNode):
    strings: list[str | nodetool.metadata.types.TextRef] | GraphNode | tuple[GraphNode, str] = Field(default=[], description=None)
    separator: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.Join"



class Replace(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    old: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    new: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.Replace"



class SaveText(GraphNode):
    value: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description=None)
    name: str | GraphNode | tuple[GraphNode, str] = Field(default='text.txt', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.SaveText"



class Split(GraphNode):
    text: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    delimiter: str | GraphNode | tuple[GraphNode, str] = Field(default=',', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.Split"



class Template(GraphNode):
    string: str | nodetool.metadata.types.TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    values: dict[str, Any] | GraphNode | tuple[GraphNode, str] = Field(default={}, description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.Template"



class TextID(GraphNode):
    text: TextRef | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    @classmethod
    def get_node_type(cls): return "nodetool.text.TextID"





from nodetool.api.types.graph import Edge
from nodetool.dsl.graph import GraphNode
from nodetool.workflows.base_node import BaseNode, get_node_class


class GraphNodeConverter:
    """
    Converts GraphNode objects into node instances and manages the creation of edges between nodes.
    """

    edges: list[Edge]
    nodes: dict[str, BaseNode]
    next_node_id: str
    next_edge_id: str

    def __init__(self):
        self.edges = []
        self.nodes = {}
        self.next_node_id = "0"
        self.next_edge_id = "0"

    def get_next_node_id(self) -> str:
        """
        Returns the next available node ID.
        """
        self.next_node_id = str(int(self.next_node_id) + 1)
        return self.next_node_id

    def get_next_edge_id(self) -> str:
        """
        Returns the next available edge ID.
        """
        self.next_edge_id = str(int(self.next_edge_id) + 1)
        return self.next_edge_id

    def add(self, graph_node: GraphNode):
        """
        Adds a GraphNode to the converter and creates edges between nodes if necessary.

        Args:
          graph_node (GraphNode): The GraphNode to add.

        Returns:
          BaseNode: The created node instance.
        """
        # make sure all node types are imported
        import nodetool.nodes

        if graph_node.id in self.nodes:
            return self.nodes[graph_node.id]
        graph_node.id = self.get_next_node_id()
        node_data = {}
        for field_name in graph_node.model_fields.keys():
            value = getattr(graph_node, field_name)
            if value is None:
                continue
            if isinstance(value, tuple) and len(value) == 2:
                source_node, slot = value
                source_node_instance = self.add(source_node)
                self.edges.append(
                    Edge(
                        source=str(source_node_instance._id),
                        sourceHandle=slot,
                        target=str(graph_node.id),
                        targetHandle=field_name,
                    )
                )
            elif isinstance(value, GraphNode):
                source_node_instance = self.add(value)
                self.edges.append(
                    Edge(
                        source=str(source_node_instance._id),
                        sourceHandle="output",
                        target=str(graph_node.id),
                        targetHandle=field_name,
                    )
                )
            else:
                node_data[field_name] = value

        node_cls = get_node_class(graph_node.get_node_type())

        assert (
            node_cls is not None
        ), f"Node class {graph_node.get_node_type()} not found"

        node_instance = node_cls(**node_data)
        self.nodes[graph_node.id] = node_instance
        return node_instance



dependencies = ["torch"]

import torch

from custom_midas_repo.midas.dpt_depth import DPTDepthModel
from custom_midas_repo.midas.midas_net import MidasNet
from custom_midas_repo.midas.midas_net_custom import MidasNet_small

def DPT_BEiT_L_512(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_BEiT_L_512 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="beitl16_512",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_beit_large_512.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_BEiT_L_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_BEiT_L_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="beitl16_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_beit_large_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_BEiT_B_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_BEiT_B_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="beitb16_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_beit_base_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_SwinV2_L_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_SwinV2_L_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="swin2l24_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_swin2_large_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_SwinV2_B_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_SwinV2_B_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="swin2b24_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_swin2_base_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_SwinV2_T_256(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_SwinV2_T_256 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="swin2t16_256",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_swin2_tiny_256.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_Swin_L_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_Swin_L_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="swinl12_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_swin_large_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_Next_ViT_L_384(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_Next_ViT_L_384 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="next_vit_large_6m",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_next_vit_large_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_LeViT_224(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT_LeViT_224 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="levit_384",
            non_negative=True,
            head_features_1=64,
            head_features_2=8,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3_1/dpt_levit_224.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def DPT_Large(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT-Large model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="vitl16_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3/dpt_large_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model
    
def DPT_Hybrid(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS DPT-Hybrid model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = DPTDepthModel(
            path=None,
            backbone="vitb_rn50_384",
            non_negative=True,
        )

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v3/dpt_hybrid_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model
    
def MiDaS(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS v2.1 model for monocular depth estimation
    pretrained (bool): load pretrained weights into model
    """

    model = MidasNet()

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v2_1/midas_v21_384.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model

def MiDaS_small(pretrained=True, **kwargs):
    """ # This docstring shows up in hub.help()
    MiDaS v2.1 small model for monocular depth estimation on resource-constrained devices
    pretrained (bool): load pretrained weights into model
    """

    model = MidasNet_small(None, features=64, backbone="efficientnet_lite3", exportable=True, non_negative=True, blocks={'expand': True})

    if pretrained:
        checkpoint = (
            "https://github.com/isl-org/MiDaS/releases/download/v2_1/midas_v21_small_256.pt"
        )
        state_dict = torch.hub.load_state_dict_from_url(
            checkpoint, map_location=torch.device('cpu'), progress=True, check_hash=True
        )
        model.load_state_dict(state_dict)

    return model


def transforms():
    import cv2
    from torchvision.transforms import Compose
    from custom_midas_repo.midas.transforms import Resize, NormalizeImage, PrepareForNet
    from custom_midas_repo.midas import transforms

    transforms.default_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                384,
                384,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method="upper_bound",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.small_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                256,
                256,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method="upper_bound",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.dpt_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                384,
                384,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method="minimal",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.beit512_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                512,
                512,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method="minimal",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.swin384_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                384,
                384,
                resize_target=None,
                keep_aspect_ratio=False,
                ensure_multiple_of=32,
                resize_method="minimal",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.swin256_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                256,
                256,
                resize_target=None,
                keep_aspect_ratio=False,
                ensure_multiple_of=32,
                resize_method="minimal",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    transforms.levit_transform = Compose(
        [
            lambda img: {"image": img / 255.0},
            Resize(
                224,
                224,
                resize_target=None,
                keep_aspect_ratio=False,
                ensure_multiple_of=32,
                resize_method="minimal",
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
            PrepareForNet(),
            lambda sample: torch.from_numpy(sample["image"]).unsqueeze(0),
        ]
    )

    return transforms



from ..utils import common_annotator_call, create_node_input_types
import comfy.model_management as model_management

class Scribble_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types()

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, resolution=512, **kwargs):
        from controlnet_aux.scribble import ScribbleDetector

        model = ScribbleDetector()
        return (common_annotator_call(model, image, resolution=resolution), )

class Scribble_XDoG_Preprocessor:
    @classmethod
    def INPUT_TYPES(s):
        return create_node_input_types(
            threshold = ("INT", {"default": 32, "min": 1, "max": 64, "step": 64})
        )

    RETURN_TYPES = ("IMAGE",)
    FUNCTION = "execute"

    CATEGORY = "ControlNet Preprocessors/Line Extractors"

    def execute(self, image, resolution=512, **kwargs):
        from controlnet_aux.scribble import ScribbleXDog_Detector

        model = ScribbleXDog_Detector()
        return (common_annotator_call(model, image, resolution=resolution), )

NODE_CLASS_MAPPINGS = {
    "ScribblePreprocessor": Scribble_Preprocessor,
    "Scribble_XDoG_Preprocessor": Scribble_XDoG_Preprocessor
}
NODE_DISPLAY_NAME_MAPPINGS = {
    "ScribblePreprocessor": "Scribble Lines",
    "Scribble_XDoG_Preprocessor": "Scribble XDoG Lines"
}


import json
import logging
import os
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union

import torch
from torch.hub import HASH_REGEX, download_url_to_file, urlparse

try:
    from torch.hub import get_dir
except ImportError:
    from torch.hub import _get_torch_home as get_dir

from custom_timm import __version__

try:
    from huggingface_hub import (create_repo, get_hf_file_metadata,
                                 hf_hub_download, hf_hub_url,
                                 repo_type_and_id_from_hf_id, upload_folder)
    from huggingface_hub.utils import EntryNotFoundError
    hf_hub_download = partial(hf_hub_download, library_name="timm", library_version=__version__)
    _has_hf_hub = True
except ImportError:
    hf_hub_download = None
    _has_hf_hub = False

_logger = logging.getLogger(__name__)


def get_cache_dir(child_dir=''):
    """
    Returns the location of the directory where models are cached (and creates it if necessary).
    """
    # Issue warning to move data if old env is set
    if os.getenv('TORCH_MODEL_ZOO'):
        _logger.warning('TORCH_MODEL_ZOO is deprecated, please use env TORCH_HOME instead')

    hub_dir = get_dir()
    child_dir = () if not child_dir else (child_dir,)
    model_dir = os.path.join(hub_dir, 'checkpoints', *child_dir)
    os.makedirs(model_dir, exist_ok=True)
    return model_dir


def download_cached_file(url, check_hash=True, progress=False):
    parts = urlparse(url)
    filename = os.path.basename(parts.path)
    cached_file = os.path.join(get_cache_dir(), filename)
    if not os.path.exists(cached_file):
        _logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
        hash_prefix = None
        if check_hash:
            r = HASH_REGEX.search(filename)  # r is Optional[Match[str]]
            hash_prefix = r.group(1) if r else None
        download_url_to_file(url, cached_file, hash_prefix, progress=progress)
    return cached_file


def has_hf_hub(necessary=False):
    if not _has_hf_hub and necessary:
        # if no HF Hub module installed, and it is necessary to continue, raise error
        raise RuntimeError(
            'Hugging Face hub model specified but package not installed. Run `pip install huggingface_hub`.')
    return _has_hf_hub


def hf_split(hf_id):
    # FIXME I may change @ -> # and be parsed as fragment in a URI model name scheme
    rev_split = hf_id.split('@')
    assert 0 < len(rev_split) <= 2, 'hf_hub id should only contain one @ character to identify revision.'
    hf_model_id = rev_split[0]
    hf_revision = rev_split[-1] if len(rev_split) > 1 else None
    return hf_model_id, hf_revision


def load_cfg_from_json(json_file: Union[str, os.PathLike]):
    with open(json_file, "r", encoding="utf-8") as reader:
        text = reader.read()
    return json.loads(text)


def _download_from_hf(model_id: str, filename: str):
    hf_model_id, hf_revision = hf_split(model_id)
    return hf_hub_download(hf_model_id, filename, revision=hf_revision)


def load_model_config_from_hf(model_id: str):
    assert has_hf_hub(True)
    cached_file = _download_from_hf(model_id, 'config.json')
    pretrained_cfg = load_cfg_from_json(cached_file)
    pretrained_cfg['hf_hub_id'] = model_id  # insert hf_hub id for pretrained weight load during model creation
    pretrained_cfg['source'] = 'hf-hub'
    model_name = pretrained_cfg.get('architecture')
    return pretrained_cfg, model_name


def load_state_dict_from_hf(model_id: str, filename: str = 'pytorch_model.bin'):
    assert has_hf_hub(True)
    cached_file = _download_from_hf(model_id, filename)
    state_dict = torch.load(cached_file, map_location='cpu')
    return state_dict


def save_for_hf(model, save_directory, model_config=None):
    assert has_hf_hub(True)
    model_config = model_config or {}
    save_directory = Path(save_directory)
    save_directory.mkdir(exist_ok=True, parents=True)

    weights_path = save_directory / 'pytorch_model.bin'
    torch.save(model.state_dict(), weights_path)

    config_path = save_directory / 'config.json'
    hf_config = model.pretrained_cfg
    hf_config['num_classes'] = model_config.pop('num_classes', model.num_classes)
    hf_config['num_features'] = model_config.pop('num_features', model.num_features)
    hf_config['labels'] = model_config.pop('labels', [f"LABEL_{i}" for i in range(hf_config['num_classes'])])
    hf_config.update(model_config)

    with config_path.open('w') as f:
        json.dump(hf_config, f, indent=2)


def push_to_hf_hub(
    model,
    repo_id: str,
    commit_message: str ='Add model',
    token: Optional[str] = None,
    revision: Optional[str] = None,
    private: bool = False,
    create_pr: bool = False,
    model_config: Optional[dict] = None,
):
    # Create repo if doesn't exist yet
    repo_url = create_repo(repo_id, token=token, private=private, exist_ok=True)

    # Infer complete repo_id from repo_url
    # Can be different from the input `repo_id` if repo_owner was implicit
    _, repo_owner, repo_name = repo_type_and_id_from_hf_id(repo_url)
    repo_id = f"{repo_owner}/{repo_name}"

    # Check if README file already exist in repo
    try:
        get_hf_file_metadata(hf_hub_url(repo_id=repo_id, filename="README.md", revision=revision))
        has_readme = True
    except EntryNotFoundError:
        has_readme = False

    # Dump model and push to Hub
    with TemporaryDirectory() as tmpdir:
        # Save model weights and config.
        save_for_hf(model, tmpdir, model_config=model_config)

        # Add readme if does not exist
        if not has_readme:
            readme_path = Path(tmpdir) / "README.md"
            readme_text = f'---\ntags:\n- image-classification\n- timm\nlibrary_tag: timm\n---\n# Model card for {repo_id}'
            readme_path.write_text(readme_text)

        # Upload model and return
        return upload_folder(
            repo_id=repo_id,
            folder_path=tmpdir,
            revision=revision,
            create_pr=create_pr,
            commit_message=commit_message,
        )



from nodetool.common.replicate_node import replicate_node
from nodetool.metadata.types import AudioRef, ImageRef, VideoRef
import argparse
import dotenv

"""
This script generates source code for all replicate nodes 
using information from Replicate's model API.
"""

replicate_nodes = [
    {
        "node_name": "AdInpaint",
        "namespace": "image.generate",
        "model_id": "logerzhu/ad-inpaint",
        "return_type": ImageRef,
        "overrides": {"image_path": ImageRef},
    },
    {
        "node_name": "ConsistentCharacter",
        "model_id": "fofr/consistent-character",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"subject": ImageRef},
    },
    {
        "node_name": "PulidBase",
        "model_id": "fofr/pulid-base",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"face_image": ImageRef},
    },
    {
        "model_id": "lucataco/proteus-v0.4",
        "node_name": "Proteus",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "node_name": "SDXLClipInterrogator",
        "namespace": "image.analyze",
        "model_id": "lucataco/sdxl-clip-interrogator",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "lucataco/moondream2",
        "namespace": "image.analyze",
        "node_name": "Moondream2",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "abiruyt/text-extract-ocr",
        "node_name": "TextExtractOCR",
        "namespace": "image.ocr",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "mickeybeurskens/latex-ocr",
        "node_name": "LatexOCR",
        "namespace": "image.ocr",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "fofr/face-to-many",
        "node_name": "FaceToMany",
        "namespace": "image.face",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "BecomeImage",
        "namespace": "image.face",
        "model_id": "fofr/become-image",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "image_to_become": ImageRef},
    },
    {
        "model_id": "tencentarc/photomaker",
        "node_name": "PhotoMaker",
        "namespace": "image.face",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "tencentarc/photomaker-style",
        "node_name": "PhotoMakerStyle",
        "namespace": "image.face",
        "return_type": ImageRef,
        "overrides": {
            "input_image": ImageRef,
            "input_image2": ImageRef,
            "input_image3": ImageRef,
            "input_image4": ImageRef,
        },
    },
    {
        "model_id": "fofr/face-to-sticker",
        "node_name": "FaceToSticker",
        "namespace": "image.face",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "zsxkib/instant-id",
        "node_name": "InstantId",
        "namespace": "image.face",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "pose_image": ImageRef},
    },
    {
        "node_name": "RealEsrGan",
        "namespace": "image.upscale",
        "model_id": "daanelson/real-esrgan-a100",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "ClarityUpscaler",
        "namespace": "image.upscale",
        "model_id": "philz1337x/clarity-upscaler",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "MagicImageRefiner",
        "namespace": "image.upscale",
        "model_id": "batouresearch/magic-image-refiner",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "node_name": "ruDallE_SR",
        "namespace": "image.upscale",
        "model_id": "cjwbw/rudalle-sr",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "HighResolutionControlNetTile",
        "namespace": "image.upscale",
        "model_id": "batouresearch/high-resolution-controlnet-tile",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "UltimateSDUpscale",
        "namespace": "image.upscale",
        "model_id": "fewjative/ultimate-sd-upscale",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "Maxim",
        "model_id": "google-research/maxim",
        "namespace": "image.enhance",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "SwinIR",
        "namespace": "image.upscale",
        "model_id": "jingyunliang/swinir",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "Swin2SR",
        "namespace": "image.upscale",
        "model_id": "mv-lab/swin2sr",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "stability-ai/stable-diffusion",
        "node_name": "StableDiffusion",
        "namespace": "image.generate",
        "return_type": ImageRef,
    },
    {
        "model_id": "usamaehsan/controlnet-1.1-x-realistic-vision-v2.0",
        "node_name": "Controlnet_Realistic_Vision",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "stability-ai/sdxl",
        "node_name": "StableDiffusionXL",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "lucataco/juggernaut-xl-v9",
        "node_name": "Juggernaut_XL_V9",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "fofr/epicrealismxl-lightning-hades",
        "node_name": "EpicRealismXL_Lightning_Hades",
        "namespace": "image.generate",
        "return_type": ImageRef,
    },
    {
        "model_id": "swartype/sdxl-pixar",
        "node_name": "SDXL_Pixar",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "fofr/sdxl-emoji",
        "node_name": "SDXL_Emoji",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "lucataco/sdxl-inpainting",
        "node_name": "StableDiffusionInpainting",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "lucataco/realvisxl-v2.0",
        "node_name": "RealVisXL_V2",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "lucataco/realvisxl2-lcm",
        "node_name": "RealVisXL2_LCM",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef, "mask": ImageRef},
    },
    {
        "model_id": "fofr/realvisxl-v3-multi-controlnet-lora",
        "node_name": "RealVisXL_V3_Multi_Controlnet_Lora",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {
            "image": ImageRef,
            "mask": ImageRef,
            "controlnet_1_image": ImageRef,
            "controlnet_2_image": ImageRef,
            "controlnet_3_image": ImageRef,
        },
    },
    {
        "model_id": "batouresearch/open-dalle-1.1-lora",
        "node_name": "OpenDalle_Lora",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {
            "image": ImageRef,
            "mask": ImageRef,
        },
    },
    {
        "model_id": "usamaehsan/controlnet-x-ip-adapter-realistic-vision-v5",
        "node_name": "Controlnet_X_IP_Adapter_Realistic_Vision_V5",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {
            "image": ImageRef,
            "lineart_image": ImageRef,
            "scribble_image": ImageRef,
            "tile_image": ImageRef,
            "brightness_image": ImageRef,
            "inpainting_image": ImageRef,
            "mask_image": ImageRef,
        },
    },
    {
        "model_id": "lucataco/sdxl-controlnet",
        "node_name": "SDXL_Controlnet",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "catacolabs/sdxl-ad-inpaint",
        "node_name": "SDXL_Ad_Inpaint",
        "namespace": "image.generate",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "OldPhotosRestoration",
        "namespace": "image.enhance",
        "model_id": "microsoft/bringing-old-photos-back-to-life",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "Kandinsky",
        "namespace": "image.generate",
        "model_id": "ai-forever/kandinsky-2.2",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "StableDiffusionXLLightning",
        "namespace": "image.generate",
        "model_id": "bytedance/sdxl-lightning-4step",
        "return_type": ImageRef,
    },
    {
        "node_name": "PlaygroundV2",
        "namespace": "image.generate",
        "model_id": "playgroundai/playground-v2.5-1024px-aesthetic",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "VideoMorpher",
        "model_id": "fofr/video-morpher",
        "return_type": VideoRef,
        "namespace": "video.generate",
        "overrides": {
            "subject_image_1": ImageRef,
            "subject_image_2": ImageRef,
            "subject_image_3": ImageRef,
            "subject_image_4": ImageRef,
            "style_image": ImageRef,
        },
    },
    {
        "node_name": "StyleTransfer",
        "model_id": "fofr/style-transfer",
        "return_type": ImageRef,
        "namespace": "image.generate",
        "overrides": {"structure_image": ImageRef, "style_image": ImageRef},
    },
    {
        "node_name": "HotshotXL",
        "namespace": "video.generate",
        "model_id": "lucataco/hotshot-xl",
        "return_type": VideoRef,
    },
    {
        "node_name": "AnimateDiff",
        "namespace": "video.generate",
        "model_id": "zsxkib/animate-diff",
        "return_type": VideoRef,
    },
    {
        "node_name": "Tooncrafter",
        "namespace": "video.generate",
        "model_id": "fofr/tooncrafter",
        "return_type": VideoRef,
        "overrides": {
            "image_1": ImageRef,
            "image_2": ImageRef,
            "image_3": ImageRef,
            "image_4": ImageRef,
            "image_5": ImageRef,
            "image_6": ImageRef,
            "image_7": ImageRef,
            "image_8": ImageRef,
            "image_9": ImageRef,
            "image_10": ImageRef,
        },
    },
    {
        "node_name": "Zeroscope_V2_XL",
        "namespace": "video.generate",
        "model_id": "anotherjesse/zeroscope-v2-xl",
        "return_type": VideoRef,
    },
    {
        "node_name": "RobustVideoMatting",
        "namespace": "video.generate",
        "model_id": "arielreplicate/robust_video_matting",
        "return_type": VideoRef,
        "overrides": {"input_video": VideoRef},
    },
    {
        "node_name": "StableDiffusionInfiniteZoom",
        "namespace": "video.generate",
        "model_id": "arielreplicate/stable_diffusion_infinite_zoom",
        "return_type": VideoRef,
    },
    {
        "node_name": "AnimateDiffIllusions",
        "namespace": "video.generate",
        "model_id": "zsxkib/animatediff-illusions",
        "return_type": VideoRef,
        "overrides": {"controlnet_video": VideoRef},
    },
    {
        "node_name": "Illusions",
        "namespace": "image.generate",
        "model_id": "fofr/illusions",
        "return_type": ImageRef,
        "overrides": {
            "image": ImageRef,
            "control_image": ImageRef,
            "mask_image": ImageRef,
        },
    },
    {
        "model_id": "daanelson/minigpt-4",
        "node_name": "MiniGPT4",
        "namespace": "image.analyze",
        "return_type": "str",
    },
    {
        "model_id": "lucataco/nsfw_image_detection",
        "node_name": "NSFWImageDetection",
        "namespace": "image.analyze",
        "return_type": "str",
    },
    {
        "model_id": "yorickvp/llava-v1.6-34b",
        "node_name": "Llava34B",
        "namespace": "image.analyze",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "salesforce/blip",
        "node_name": "Blip",
        "namespace": "image.analyze",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "andreasjansson/blip-2",
        "node_name": "Blip2",
        "namespace": "image.analyze",
        "return_type": "str",
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "pharmapsychotic/clip-interrogator",
        "node_name": "ClipInterrogator",
        "namespace": "image.analyze",
        "return_type": str,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "Llava13b",
        "namespace": "image.analyze",
        "model_id": "yorickvp/llava-13b",
        "return_type": str,
        "overrides": {"image": ImageRef},
    },
    {
        "model_id": "meta/meta-llama-3-8b",
        "node_name": "Llama3_8B",
        "namespace": "text.generate",
        "return_type": str,
    },
    {
        "model_id": "meta/meta-llama-3-70b",
        "node_name": "Llama3_70B",
        "namespace": "text.generate",
        "return_type": str,
    },
    {
        "model_id": "meta/meta-llama-3-8b-instruct",
        "node_name": "Llama3_8B_Instruct",
        "namespace": "text.generate",
        "return_type": str,
    },
    {
        "model_id": "meta/meta-llama-3-70b-instruct",
        "node_name": "Llama3_70B_Instruct",
        "namespace": "text.generate",
        "return_type": str,
    },
    {
        "model_id": "snowflake/snowflake-arctic-instruct",
        "node_name": "Snowflake_Arctic_Instruct",
        "namespace": "text.generate",
        "return_type": str,
    },
    {
        "model_id": "ryan5453/demucs",
        "node_name": "Demucs",
        "namespace": "audio.separate",
        "overrides": {"audio": AudioRef},
        "return_type": {
            "vocals": AudioRef,
            "drums": AudioRef,
            "bass": AudioRef,
            "other": AudioRef,
        },
    },
    {
        "model_id": "openai/whisper",
        "node_name": "Whisper",
        "namespace": "audio.transcribe",
        "return_type": str,
        "overrides": {"audio": AudioRef},
    },
    {
        "model_id": "vaibhavs10/incredibly-fast-whisper",
        "node_name": "IncrediblyFastWhisper",
        "namespace": "audio.transcribe",
        "return_type": str,
        "overrides": {"audio": AudioRef},
    },
    {
        "node_name": "AudioSuperResolution",
        "namespace": "audio.enhance",
        "model_id": "nateraw/audio-super-resolution",
        "return_type": AudioRef,
        "overrides": {"input_file": AudioRef},
    },
    {
        "node_name": "RemoveBackground",
        "namespace": "image.process",
        "model_id": "cjwbw/rembg",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "ModNet",
        "namespace": "image.process",
        "model_id": "pollinations/modnet",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "CodeFormer",
        "namespace": "image.enhance",
        "model_id": "lucataco/codeformer",
        "return_type": ImageRef,
        "overrides": {"image": ImageRef},
    },
    {
        "node_name": "RealisticVoiceCloning",
        "namespace": "audio.generate",
        "model_id": "zsxkib/realistic-voice-cloning",
        "return_type": AudioRef,
        "overrides": {"song_input": AudioRef},
    },
    {
        "node_name": "TortoiseTTS",
        "model_id": "afiaka87/tortoise-tts",
        "namespace": "audio.generate",
        "return_type": AudioRef,
        "overrides": {"custom_voice": AudioRef},
    },
    {
        "model_id": "adirik/styletts2",
        "node_name": "StyleTTS2",
        "namespace": "audio.generate",
        "return_type": AudioRef,
        "overrides": {"reference": AudioRef},
    },
    {
        "node_name": "Riffusion",
        "namespace": "audio.generate",
        "model_id": "riffusion/riffusion",
        "return_type": AudioRef,
        "output_key": "audio",
        "overrides": {"song_input": AudioRef},
    },
    {
        "node_name": "Bark",
        "namespace": "audio.generate",
        "model_id": "suno-ai/bark",
        "return_type": AudioRef,
        "output_key": "audio_out",
    },
    {
        "node_name": "MusicGen",
        "namespace": "audio.generate",
        "model_id": "meta/musicgen",
        "return_type": AudioRef,
    },
    {
        "node_name": "VideoLlava",
        "namespace": "video.analyze",
        "model_id": "nateraw/video-llava",
        "return_type": str,
        "overrides": {"video_path": VideoRef, "image_path": ImageRef},
    },
    {
        "node_name": "AudioToWaveform",
        "namespace": "video.generate",
        "model_id": "fofr/audio-to-waveform",
        "return_type": VideoRef,
        "overrides": {"audio": AudioRef},
    },
]

if __name__ == "__main__":
    dotenv.load_dotenv()

    parser = argparse.ArgumentParser()
    parser.add_argument("--namespace", help="Specify the namespace argument")
    args = parser.parse_args()

    # Use the namespace argument here
    if args.namespace:
        for node in replicate_nodes:
            if node["namespace"] == args.namespace:
                replicate_node(**node)
    else:
        for node in replicate_nodes:
            replicate_node(**node)



#https://github.com/siyeong0/Anime-Face-Segmentation/blob/main/util.py
#The color palette is changed according to https://github.com/Mikubill/sd-webui-controlnet/blob/91f67ddcc7bc47537a6285864abfc12590f46c3f/annotator/anime_face_segment/__init__.py
import cv2 as cv
import glob
import numpy as np
import os

"""
COLOR_BACKGROUND = (0,255,255)
COLOR_HAIR = (255,0,0)
COLOR_EYE = (0,0,255)
COLOR_MOUTH = (255,255,255)
COLOR_FACE = (0,255,0)
COLOR_SKIN = (255,255,0)
COLOR_CLOTHES = (255,0,255)
"""
COLOR_BACKGROUND = (255,255,0)
COLOR_HAIR = (0,0,255)
COLOR_EYE = (255,0,0)
COLOR_MOUTH = (255,255,255)
COLOR_FACE = (0,255,0)
COLOR_SKIN = (0,255,255)
COLOR_CLOTHES = (255,0,255)
PALETTE = [COLOR_BACKGROUND,COLOR_HAIR,COLOR_EYE,COLOR_MOUTH,COLOR_FACE,COLOR_SKIN,COLOR_CLOTHES]

def img2seg(path):
    src = cv.imread(path)
    src = src.reshape(-1, 3)
    seg_list = []
    for color in PALETTE:
        seg_list.append(np.where(np.all(src==color, axis=1), 1.0, 0.0))
    dst = np.stack(seg_list,axis=1).reshape(512,512,7)
    
    return dst.astype(np.float32)

def seg2img(src):
    src = np.moveaxis(src,0,2)
    dst = [[PALETTE[np.argmax(val)] for val in buf]for buf in src]
    
    return np.array(dst).astype(np.uint8)


from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.huggingface.audio.generate import TextToAudioModel

class TextToAudio(GraphNode):
    model: TextToAudioModel | GraphNode | tuple[GraphNode, str] = Field(default=TextToAudioModel('facebook/musicgen-small'), description='The model ID to use for the audio generation')
    inputs: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='The input text to the model')
    @classmethod
    def get_node_type(cls): return "huggingface.audio.generate.TextToAudio"


from nodetool.nodes.huggingface.audio.generate import ModelId

class TextToSpeech(GraphNode):
    model: ModelId | GraphNode | tuple[GraphNode, str] = Field(default=ModelId('suno/bark'), description='The model ID to use for the image generation')
    inputs: str | GraphNode | tuple[GraphNode, str] = Field(default=PydanticUndefined, description='The input text to the model')
    @classmethod
    def get_node_type(cls): return "huggingface.audio.generate.TextToSpeech"





import io
from enum import Enum
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
from io import StringIO
from pydantic import Field
from sklearn.datasets import load_iris
from sklearn.utils import Bunch
from nodetool.metadata.types import Tensor
from nodetool.metadata.types import FolderRef
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.metadata.types import DataframeRef
from nodetool.metadata.types import ImageRef
from nodetool.workflows.base_node import BaseNode
from typing import Any, Literal
from nodetool.metadata.types import Dataset


class SaveDataframe(BaseNode):
    """
    Save data frame as a CSV file in a folder.
    csv, folder
    """

    df: DataframeRef = DataframeRef()
    folder: FolderRef = Field(
        default=FolderRef(), description="Name of the output folder."
    )
    name: str = Field(default="output.csv", description="Name of the output file.")

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.df)
        return await context.dataframe_from_pandas(df, self.name)


class SelectColumn(BaseNode):
    """
    Select specific columns from a dataframe.
    dataframe, columns
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(),
        description="a dataframe from which columns are to be selected",
    )
    columns: str = Field("", description="comma separated list of column names")

    async def process(self, context: ProcessingContext) -> DataframeRef:
        columns = self.columns.split(",")
        df = await context.dataframe_to_pandas(self.dataframe)
        return await context.dataframe_from_pandas(df[columns])  # type: ignore


class ColumnToList(BaseNode):
    """
    Convert a column in a dataframe to a list.
    dataframe, column
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )
    column_name: str = Field(
        default="", description="The name of the column to be converted to a list."
    )

    async def process(self, context: ProcessingContext) -> list[Any]:
        df = await context.dataframe_to_pandas(self.dataframe)
        return df[self.column_name].tolist()


class RowsToStrings(BaseNode):
    """
    Convert rows in a dataframe to strings.
    dataframe, columns
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )
    template: str = Field(
        default="",
        description="The template for the string representation. Each column can be referenced by {column_name}.",
    )

    async def process(self, context: ProcessingContext) -> list[str]:
        df = await context.dataframe_to_pandas(self.dataframe)
        return [self.template.format(**row) for _, row in df.iterrows()]


class ListToColumn(BaseNode):
    """
    Convert a list of objects into a column in a dataframe.
    dataframe, column, values
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(),
        description="Dataframe object to add a new column to.",
    )
    column_name: str = Field(
        default="",
        description="The name of the new column to be added to the dataframe.",
    )
    values: list[Any] = Field(
        default_factory=list,
        description="A list of any type of elements which will be the new column's values.",
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.dataframe)
        df[self.column_name] = self.values
        return await context.dataframe_from_pandas(df)


class ListToDataFrame(BaseNode):
    """
    Convert a list of values to a dataframe. Each row may be a dict, list, or single value.
    list, dataframe
    """

    values: list[Any] = Field(
        title="Values",
        default=[],
        description="List of values to be converted, each value will be a row.",
    )
    columns: str = Field(
        title="Columns", default="", description="Comma separated list of column names"
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        rows = []
        columns = self.columns.split(",")
        for value in self.values:
            if not isinstance(value, dict):
                raise ValueError("List must contain dicts.")
            if type(value) == dict:
                row = {}
                for k, v in value.items():
                    if type(v) == dict:
                        row[k] = v["value"]
                    elif type(v) in [int, float, str, bool]:
                        row[k] = v
                    else:
                        row[k] = str(v)
            elif type(value) in [int, float, str, bool]:
                col = columns[0] if len(columns) > 0 else "value"
                row = {col: value}
            elif type(value) in [list, tuple]:
                row = {}
                for i, v in enumerate(value):
                    col = columns[i] if i < len(columns) else f"col{i}"
                    row[col] = v
            else:
                col = columns[0] if len(columns) > 0 else "value"
                row = {col: str(value)}

            rows.append(row)
        df = pd.DataFrame(rows)
        return await context.dataframe_from_pandas(df)


class CSVToDataframe(BaseNode):
    """
    Converts CSV data into a DataFrame.
    csv, dataframe
    """

    csv_data: str = Field(
        default="", title="CSV Data", description="String input of CSV formatted text."
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = pd.read_csv(StringIO(self.csv_data))
        return await context.dataframe_from_pandas(df)


class Concat(BaseNode):
    """
    Merge two dataframes along their columns.
    merge
    Outputs a single DataFrame resulting from the merging of dataframe_a and dataframe_b along their columns.
    """

    dataframe_a: DataframeRef = Field(
        default=DataframeRef(), description="First DataFrame to be merged."
    )
    dataframe_b: DataframeRef = Field(
        default=DataframeRef(), description="Second DataFrame to be merged."
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df_a = await context.dataframe_to_pandas(self.dataframe_a)
        df_b = await context.dataframe_to_pandas(self.dataframe_b)
        df = pd.concat([df_a, df_b], axis=1)
        return await context.dataframe_from_pandas(df)


class Append(BaseNode):
    """
    Append two dataframes along their rows.
    row, concat, merge
    Outputs a DataFrame object that is the result of appending DataFrame A and DataFrame B.
    """

    dataframe_a: DataframeRef = Field(
        default=DataframeRef(), description="First DataFrame to be appended."
    )
    dataframe_b: DataframeRef = Field(
        default=DataframeRef(), description="Second DataFrame to be appended."
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df_a = await context.dataframe_to_pandas(self.dataframe_a)
        df_b = await context.dataframe_to_pandas(self.dataframe_b)
        if not df_a.columns.equals(df_b.columns):
            raise ValueError(
                f"Columns in dataframe A ({df_a.columns}) do not match columns in dataframe B ({df_b.columns})"
            )
        df = pd.concat([df_a, df_b], axis=0)
        return await context.dataframe_from_pandas(df)


class Join(BaseNode):
    """
    Merges two dataframes along their columns.
    merge, join
    Outputs a single DataFrame resulting from the merging of dataframe_a and dataframe_b along their columns.
    """

    dataframe_a: DataframeRef = Field(
        default=DataframeRef(), description="First DataFrame to be merged."
    )
    dataframe_b: DataframeRef = Field(
        default=DataframeRef(), description="Second DataFrame to be merged."
    )
    join_on: str = Field(
        default="",
        description="The column name on which to join the two dataframes.",
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df_a = await context.dataframe_to_pandas(self.dataframe_a)
        df_b = await context.dataframe_to_pandas(self.dataframe_b)
        if not df_a.columns.equals(df_b.columns):
            raise ValueError(
                f"Columns in dataframe A ({df_a.columns}) do not match columns in dataframe B ({df_b.columns})"
            )
        df = pd.merge(df_a, df_b, on=self.join_on)
        return await context.dataframe_from_pandas(df)


class ToTensor(BaseNode):
    """
    Convert a dataframe to a tensor.
    dataframe, tensor
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )

    async def process(self, context: ProcessingContext) -> Tensor:
        df = await context.dataframe_to_pandas(self.dataframe)
        return Tensor.from_numpy(df.to_numpy())


class FromTensor(BaseNode):
    """
    Convert a tensor to a dataframe.
    dataframe, tensor
    """

    tensor: Tensor = Field(
        default=Tensor(),
        description="A tensor object to be converted into a dataframe.",
    )
    columns: list[str] = Field(
        default=[],
        description="A list of strings specifying the column names for the resulting dataframe.",
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        array = self.tensor.to_numpy()
        if array.shape[1] != len(self.columns):
            raise ValueError(
                f"Number of columns in tensor ({array.shape[1]}) does not match number of columns specified ({len(self.columns)})"
            )
        df = pd.DataFrame(array, columns=self.columns)
        return await context.dataframe_from_pandas(df)


class Plot(BaseNode):
    """
    Plots a dataframe as a line, bar, or scatter plot.
    plot, dataframe, line, bar, scatter
    """

    class PlotType(str, Enum):
        LINE = "line"
        BAR = "bar"
        SCATTER = "scatter"

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )
    x_column: str = Field(
        default="",
        description="The name of the x column to be used in the plot.",
    )
    y_column: str = Field(
        default="",
        description="The name of the y column to be used in the plot.",
    )
    plot_type: PlotType = Field(
        default=PlotType.LINE,
        description="The type of plot to be created. Can be 'line', 'bar', or 'scatter'.",
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        df = await context.dataframe_to_pandas(self.dataframe)
        if self.x_column not in df.columns:  # type: ignore
            raise ValueError(f"Invalid x_column: {self.x_column}")
        if self.y_column not in df.columns:  # type: ignore
            raise ValueError(f"Invalid y_column: {self.y_column}")
        if self.plot_type == self.PlotType.LINE:
            plot = sns.lineplot(x=self.x_column, y=self.y_column, data=df)
        elif self.plot_type == self.PlotType.BAR:
            plot = sns.barplot(x=self.x_column, y=self.y_column, data=df)
        elif self.plot_type == self.PlotType.SCATTER:
            plot = sns.scatterplot(x=self.x_column, y=self.y_column, data=df)
        else:
            raise ValueError(f"Invalid plot type: {self.plot_type}")
        fig = plot.get_figure()
        if fig is None:
            raise ValueError("Invalid plot")
        img_bytes = io.BytesIO()
        fig.savefig(img_bytes, format="png")
        plt.close(fig)
        return await context.image_from_bytes(img_bytes.getvalue())


class PlotHistogram(BaseNode):
    """
    Plot a histogram of a dataframe column.
    histogram, plot, dataframe
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )
    column: str = Field(title="Column", default="", description="The column to plot.")

    async def process(self, context: ProcessingContext) -> ImageRef:
        df = await context.dataframe_to_pandas(self.dataframe)
        if self.column not in df.columns:
            raise ValueError(f"Invalid column: {self.column}")
        (fig, ax) = plt.subplots()
        sns.set_theme(style="darkgrid")
        sns.histplot(df[self.column], ax=ax)  # type: ignore
        img_bytes = io.BytesIO()
        fig.savefig(img_bytes, format="png")
        plt.close(fig)
        return await context.image_from_bytes(img_bytes.getvalue())


class PlotHeatmap(BaseNode):
    """
    Plot a heatmap of a dataframe.
    heatmap, plot, dataframe
    """

    dataframe: DataframeRef = Field(
        default=DataframeRef(), description="The input dataframe."
    )

    async def process(self, context: ProcessingContext) -> ImageRef:
        df = await context.dataframe_to_pandas(self.dataframe)
        sns.set_theme(style="darkgrid")
        (fig, ax) = plt.subplots()
        sns.heatmap(df, ax=ax)
        img_bytes = io.BytesIO()
        fig.savefig(img_bytes, format="png")
        plt.close(fig)
        return await context.image_from_bytes(img_bytes.getvalue())


class Filter(BaseNode):
    """
    Filter DataFrame based on a condition.
    dataframe, condition, query
    """

    df: DataframeRef = Field(
        default=DataframeRef(), description="The DataFrame to filter."
    )
    condition: str = Field(
        default="",
        description="The filtering condition to be applied to the DataFrame, e.g. column_name > 5.",
    )

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.df)
        res = df.query(self.condition)
        return await context.dataframe_from_pandas(res)


class Sort(BaseNode):
    """
    Sort a DataFrame by a column.
    dataframe, sort, order
    """

    df: DataframeRef = Field(default=DataframeRef())
    column: str = Field(default="", description="The column to sort the DataFrame by.")

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.df)
        res = df.sort_values(self.column)
        return await context.dataframe_from_pandas(res)


class DropDuplicates(BaseNode):
    """
    Drop duplicate rows from a DataFrame.
    duplicate, dataframe, unique
    """

    df: DataframeRef = Field(default=DataframeRef(), description="The input DataFrame.")

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.df)
        res = df.drop_duplicates()
        return await context.dataframe_from_pandas(res)


class DropNA(BaseNode):
    """
    Drop NA values from a DataFrame.
    na, dataframe, missing
    """

    df: DataframeRef = Field(default=DataframeRef(), description="The input DataFrame.")

    async def process(self, context: ProcessingContext) -> DataframeRef:
        df = await context.dataframe_to_pandas(self.df)
        res = df.dropna()
        return await context.dataframe_from_pandas(res)


class IrisDataFrame(BaseNode):
    """
    Load the Iris dataset.
    ml, training, dataset, test, iris
    """

    async def process(self, context: ProcessingContext) -> Dataset:
        data = load_iris(as_frame=True)
        assert isinstance(data, Bunch)
        return Dataset(
            data=await context.dataframe_from_pandas(data.data),
            target=await context.dataframe_from_pandas(pd.DataFrame(data.target)),
        )



from nodetool.metadata.types import TypeMetadata

import annotated_types
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from typing import Any, Optional


class Property(BaseModel):
    """
    Property of a node.
    """

    name: str
    type: TypeMetadata
    default: Optional[Any] = None
    title: Optional[str] = None
    description: Optional[str] = None
    min: Optional[float] = None
    max: Optional[float] = None

    def get_json_schema(self):
        """
        Returns a JSON schema for the self.
        """
        schema = self.type.get_json_schema()
        schema["description"] = self.description
        if self.min:
            schema["minimum"] = self.min
        if self.max:
            schema["maximum"] = self.max
        return schema

    @staticmethod
    def from_field(name: str, type_: TypeMetadata, field: FieldInfo):
        metadata = {type(f): f for f in field.metadata}

        ge = metadata.get(annotated_types.Ge, None)
        le = metadata.get(annotated_types.Le, None)

        if field.title is None:
            title = name.replace("_", " ").title()
        else:
            title = field.title
        return Property(
            name=name,
            type=type_,
            default=field.default,
            title=title,
            description=field.description,
            min=ge.ge if ge is not None else None,
            max=le.le if le is not None else None,
        )



from pydantic import BaseModel, Field
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode
from nodetool.common.replicate_node import ReplicateNode
from enum import Enum


class Stem(str, Enum):
    NONE = "none"
    DRUMS = "drums"
    BASS = "bass"
    OTHER = "other"
    VOCALS = "vocals"
    GUITAR = "guitar"
    PIANO = "piano"


class Model(str, Enum):
    HTDEMUCS = "htdemucs"
    HTDEMUCS_FT = "htdemucs_ft"
    HTDEMUCS_6S = "htdemucs_6s"
    HDEMUCS_MMI = "hdemucs_mmi"
    MDX_Q = "mdx_q"
    MDX_EXTRA_Q = "mdx_extra_q"


class Clip_mode(str, Enum):
    RESCALE = "rescale"
    CLAMP = "clamp"
    NONE = "none"


class Mp3_preset(int, Enum):
    _2 = 2
    _3 = 3
    _4 = 4
    _5 = 5
    _6 = 6
    _7 = 7


class Wav_format(str, Enum):
    INT16 = "int16"
    INT24 = "int24"
    FLOAT32 = "float32"


class Output_format(str, Enum):
    MP3 = "mp3"
    FLAC = "flac"
    WAV = "wav"


class Demucs(ReplicateNode):
    """Demucs is an audio source separator created by Facebook Research."""

    @classmethod
    def replicate_model_id(cls):
        return "ryan5453/demucs:53191dee0efbfc3cbfdbab276b0dcce930705e9a4d1bb9fe1e2a7cdd33d9ca82"

    @classmethod
    def get_hardware(cls):
        return "Nvidia A40 GPU"

    @classmethod
    def model_info(cls):
        return {
            "cover_image_url": "https://tjzk.replicate.delivery/models_models_cover_image/c13b51bf-69a4-474e-ad73-d800466ca357/588ff17a-e9c9-49c7-b572-e1b189d87.png",
            "created_at": "2022-11-08T22:25:48.183283Z",
            "description": "Demucs is an audio source separator created by Facebook Research.",
            "github_url": "https://github.com/ryan5453/demucs-cog",
            "license_url": "https://github.com/ryan5453/demucs-cog/blob/main/LICENSE",
            "name": "demucs",
            "owner": "ryan5453",
            "paper_url": "https://arxiv.org/abs/2111.03600",
            "run_count": 342104,
            "url": "https://replicate.com/ryan5453/demucs",
            "visibility": "public",
            "hardware": "Nvidia A40 GPU",
        }

    @classmethod
    def return_type(cls):
        return {
            "vocals": AudioRef,
            "drums": AudioRef,
            "bass": AudioRef,
            "other": AudioRef,
        }

    jobs: int = Field(
        title="Jobs",
        description="Choose the number of parallel jobs to use for separation.",
        default=0,
    )
    stem: Stem = Field(
        description="If you just want to isolate one stem, you can choose it here.",
        default="none",
    )
    audio: AudioRef = Field(
        default=AudioRef(), description="Upload the file to be processed here."
    )
    model: Model = Field(
        description="Choose the demucs audio that proccesses your audio. The readme has more information on what to choose.",
        default="htdemucs",
    )
    split: bool = Field(
        title="Split",
        description="Choose whether or not the audio should be split into chunks.",
        default=True,
    )
    shifts: int = Field(
        title="Shifts",
        description="Choose the amount random shifts for equivariant stabilization. This performs multiple predictions with random shifts of the input and averages them, which makes it x times slower.",
        default=1,
    )
    overlap: float = Field(
        title="Overlap",
        description="Choose the amount of overlap between prediction windows.",
        default=0.25,
    )
    segment: int | None = Field(
        title="Segment",
        description="Choose the segment length to use for separation.",
        default=None,
    )
    clip_mode: Clip_mode = Field(
        description="Choose the strategy for avoiding clipping. Rescale will rescale entire signal if necessary or clamp will allow hard clipping.",
        default="rescale",
    )
    mp3_preset: Mp3_preset = Field(
        description="Choose the preset for the MP3 output. Higher is faster but worse wuality. If MP3 is not selected as the output type, this has no effect.",
        default=2,
    )
    wav_format: Wav_format = Field(
        description="Choose format for the WAV output. If WAV is not selected as the output type, this has no effect.",
        default="int24",
    )
    mp3_bitrate: int = Field(
        title="Mp3 Bitrate",
        description="Choose the bitrate for the MP3 output. Higher is better quality but larger file size. If MP3 is not selected as the output type, this has no effect.",
        default=320,
    )
    output_format: Output_format = Field(
        description="Choose the audio format you would like the result to be returned in.",
        default="mp3",
    )



from typing import Optional
from datetime import datetime
from nodetool.models.base_model import DBModel, DBField, create_time_ordered_uuid


class Task(DBModel):
    @classmethod
    def get_table_schema(cls):
        return {
            "table_name": "nodetool_tasks",
            "key_schema": {"id": "HASH"},
            "attribute_definitions": {"id": "S", "thread_id": "S"},
            "global_secondary_indexes": {
                "nodetool_task_thread_index": {"thread_id": "HASH"},
            },
        }

    id: str = DBField(hash_key=True)
    task_type: str = DBField(default="")
    user_id: str = DBField(default="")
    thread_id: str = DBField(default="")
    status: str = DBField(default="pending")
    name: str = DBField(default="")
    instructions: str = DBField(default="")
    dependencies: list[str] = DBField(default_factory=list)
    required_capabilities: list[str] = DBField(default_factory=list)
    started_at: datetime = DBField(default_factory=datetime.now)
    finished_at: datetime | None = DBField(default=None)
    error: str | None = DBField(default=None)
    result: str | None = DBField(default=None)
    cost: float | None = DBField(default=None)

    @classmethod
    def find(cls, user_id: str, task_id: str):
        task = cls.get(task_id)
        return task if task and task.user_id == user_id else None

    @classmethod
    def create(
        cls, user_id: str, thread_id: str, name: str, instructions: str, **kwargs
    ):
        return super().create(
            id=create_time_ordered_uuid(),
            user_id=user_id,
            thread_id=thread_id,
            name=name,
            instructions=instructions,
            **kwargs
        )

    @classmethod
    def paginate(
        cls,
        thread_id: Optional[str] = None,
        limit: int = 10,
        start_key: Optional[str] = None,
    ):
        return cls.query(
            condition="thread_id = :thread_id",
            values={":thread_id": thread_id},
            index="nodetool_task_thread_index",
            limit=limit,
            start_key=start_key,
        )



from typing import Tuple
import numpy as np
from pydantic import Field
from nodetool.workflows.processing_context import ProcessingContext
from nodetool.workflows.base_node import BaseNode
from nodetool.metadata.types import to_numpy
from nodetool.metadata.types import Tensor


def pad_arrays(a: np.ndarray, b: np.ndarray) -> Tuple[(np.ndarray, np.ndarray)]:
    """
    If one of the arguments is a scalar, both arguments are returned as is.
    Pads the smaller array with zeros so that both arrays are the same size.
    This is useful for operations like addition and subtraction.
    """
    if a.size == 1 or b.size == 1:
        return (a, b)
    if len(a) != len(b):
        if len(a) > len(b):
            b = np.pad(b, (0, (len(a) - len(b))), "constant")
        else:
            a = np.pad(a, (0, (len(b) - len(a))), "constant")
    return (a, b)


async def convert_output(
    context: ProcessingContext, output: np.ndarray
) -> float | int | Tensor:
    if output.size == 1:
        return output.item()
    else:
        return Tensor.from_numpy(output)


class BinaryOperation(BaseNode):
    _layout = "small"
    a: int | float | Tensor = Field(title="A", default=0.0)
    b: int | float | Tensor = Field(title="B", default=0.0)

    async def process(self, context: ProcessingContext) -> int | float | Tensor:
        a = to_numpy(self.a)
        b = to_numpy(self.b)
        if a.size > 1 and b.size > 1:
            (a, b) = pad_arrays(a, b)
        res = self.operation(a, b)
        return await convert_output(context, res)

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        raise NotImplementedError()


class Add(BinaryOperation):
    """
    Performs addition on two inputs.
    math, plus, add, addition, sum, +
    """

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.add(a, b)


class Subtract(BinaryOperation):
    """
    Subtracts the second input from the first.
    math, minus, difference, -
    """

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.subtract(a, b)


class Multiply(BinaryOperation):
    """
    Multiplies two inputs.
    math, product, times, *
    """

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.multiply(a, b)


class Divide(BinaryOperation):
    """
    Divides the first input by the second.
    math, division, arithmetic, quotient
    """

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.divide(a, b)


class Modulus(BinaryOperation):
    """
    Calculates the remainder of division of the first input by the second.
    math, mod, modulus, remainder, %, modulo, division
    """

    def operation(self, a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.mod(a, b)


class Sine(BaseNode):
    """
    Computes the sine of an input angle in radians.
    math, sine, radians, angle
    Returns a float value between -1 and 1.
    """

    _layout = "small"

    angle_rad: float | int | Tensor = Field(title="Angle (Radians)", default=0.0)

    async def process(self, context: ProcessingContext) -> float | Tensor:
        res = np.sin(to_numpy(self.angle_rad))
        return await convert_output(context, res)


class Cosine(BaseNode):
    """
    Computes the cosine of an input angle in radians.
    math, sine, radians, angle
    Returns a float value between -1 and 1.
    """

    _layout = "small"

    angle_rad: float | int | Tensor = Field(title="Angle (Radians)", default=0.0)

    async def process(self, context: ProcessingContext) -> float | Tensor:
        res = np.cos(to_numpy(self.angle_rad))
        return await convert_output(context, res)


class Power(BaseNode):
    """
    Computes the power of the base to the exponent.
    math, power, exponent
    Returns a float value between -1 and 1.
    """

    _layout = "small"

    base: float | int | Tensor = Field(title="Base", default=1.0)
    exponent: float | int | Tensor = Field(title="Exponent", default=2.0)

    async def process(self, context: ProcessingContext) -> float | int | Tensor:
        a = to_numpy(self.base)
        b = to_numpy(self.exponent)
        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
            (a, b) = pad_arrays(a, b)
        return await convert_output(context, np.power(a, b))


class SqrtTensor(BaseNode):
    """
    Calculates the square root of the input.
    math, power, sqrt, square, root
    """

    _layout = "small"

    x: int | float | Tensor = Field(title="Input", default=1.0)

    async def process(self, context: ProcessingContext) -> float | int | Tensor:
        return await convert_output(
            context, np.sqrt(to_numpy(self.x).astype(np.float32))
        )



from pydantic import BaseModel, Field
import nodetool.metadata.types
from nodetool.metadata.types import *
from nodetool.dsl.graph import GraphNode

from nodetool.nodes.replicate.image.upscale import Sd_model
from nodetool.nodes.replicate.image.upscale import Scheduler
from nodetool.nodes.replicate.image.upscale import Tiling_width
from nodetool.nodes.replicate.image.upscale import Tiling_height

class ClarityUpscaler(GraphNode):
    seed: int | GraphNode | tuple[GraphNode, str] = Field(default=1337, description='Random seed. Leave blank to randomize the seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='input image')
    prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='masterpiece, best quality, highres, <lora:more_details:0.5> <lora:SDXLrender_v2.0:1>', description='Prompt')
    dynamic: float | GraphNode | tuple[GraphNode, str] = Field(default=6, description='HDR, try from 3 - 9')
    sd_model: Sd_model | GraphNode | tuple[GraphNode, str] = Field(default='juggernaut_reborn.safetensors [338b85bc4f]', description='Stable Diffusion model checkpoint')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DPM++ 3M SDE Karras', description='scheduler')
    creativity: float | GraphNode | tuple[GraphNode, str] = Field(default=0.35, description='Creativity, try from 0.3 - 0.9')
    lora_links: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Link to a lora file you want to use in your upscaling. Multiple links possible, seperated by comma')
    downscaling: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='Downscale the image before upscaling. Can improve quality and speed for images with high resolution but lower quality')
    resemblance: float | GraphNode | tuple[GraphNode, str] = Field(default=0.6, description='Resemblance, try from 0.3 - 1.6')
    scale_factor: float | GraphNode | tuple[GraphNode, str] = Field(default=2, description='Scale factor')
    tiling_width: Tiling_width | GraphNode | tuple[GraphNode, str] = Field(default=112, description='Fractality, set lower tile width for a high Fractality')
    tiling_height: Tiling_height | GraphNode | tuple[GraphNode, str] = Field(default=144, description='Fractality, set lower tile height for a high Fractality')
    custom_sd_model: str | GraphNode | tuple[GraphNode, str] = Field(default='', description=None)
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='(worst quality, low quality, normal quality:2) JuggernautNegative-neg', description='Negative Prompt')
    num_inference_steps: int | GraphNode | tuple[GraphNode, str] = Field(default=18, description='Number of denoising steps')
    downscaling_resolution: int | GraphNode | tuple[GraphNode, str] = Field(default=768, description='Downscaling resolution')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.ClarityUpscaler"


from nodetool.nodes.replicate.image.upscale import Scheduler
from nodetool.nodes.replicate.image.upscale import Resolution

class HighResolutionControlNetTile(GraphNode):
    hdr: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='HDR improvement over the original image')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Control image for scribble controlnet')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Steps')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Prompt for the model')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DDIM', description='Choose a scheduler.')
    creativity: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='Denoising strength. 1 means total destruction of the original image')
    guess_mode: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='In this mode, the ControlNet encoder will try best to recognize the content of the input image even if you remove all prompts. The `guidance_scale` between 3.0 and 5.0 is recommended.')
    resolution: Resolution | GraphNode | tuple[GraphNode, str] = Field(default=2048, description='Image resolution')
    resemblance: float | GraphNode | tuple[GraphNode, str] = Field(default=0.5, description='Conditioning scale for controlnet')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7, description='Scale for classifier-free guidance')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='teeth, tooth, open mouth, longbody, lowres, bad anatomy, bad hands, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, mutant', description='Negative prompt')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.HighResolutionControlNetTile"


from nodetool.nodes.replicate.image.upscale import Scheduler
from nodetool.nodes.replicate.image.upscale import Resolution

class MagicImageRefiner(GraphNode):
    hdr: float | GraphNode | tuple[GraphNode, str] = Field(default=0, description='HDR improvement over the original image')
    mask: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='When provided, refines some section of the image. Must be the same size as the image')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Seed')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Image to refine')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Steps')
    prompt: str | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Prompt for the model')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='DDIM', description='Choose a scheduler.')
    creativity: float | GraphNode | tuple[GraphNode, str] = Field(default=0.25, description='Denoising strength. 1 means total destruction of the original image')
    guess_mode: bool | GraphNode | tuple[GraphNode, str] = Field(default=False, description='In this mode, the ControlNet encoder will try best to recognize the content of the input image even if you remove all prompts. The `guidance_scale` between 3.0 and 5.0 is recommended.')
    resolution: Resolution | GraphNode | tuple[GraphNode, str] = Field(default='original', description='Image resolution')
    resemblance: float | GraphNode | tuple[GraphNode, str] = Field(default=0.75, description='Conditioning scale for controlnet')
    guidance_scale: float | GraphNode | tuple[GraphNode, str] = Field(default=7, description='Scale for classifier-free guidance')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='teeth, tooth, open mouth, longbody, lowres, bad anatomy, bad hands, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, mutant', description='Negative prompt')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.MagicImageRefiner"



class RealEsrGan(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    scale: float | GraphNode | tuple[GraphNode, str] = Field(default=4, description='Factor to scale image by')
    face_enhance: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Face enhance')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.RealEsrGan"


from nodetool.nodes.replicate.image.upscale import Task

class Swin2SR(GraphNode):
    task: Task | GraphNode | tuple[GraphNode, str] = Field(default='real_sr', description='Choose a task')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.Swin2SR"


from nodetool.nodes.replicate.image.upscale import Noise
from nodetool.nodes.replicate.image.upscale import Task_type

class SwinIR(GraphNode):
    jpeg: int | GraphNode | tuple[GraphNode, str] = Field(default=40, description='scale factor, activated for JPEG Compression Artifact Reduction. Leave it as default or arbitrary if other tasks are selected')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='input image')
    noise: Noise | GraphNode | tuple[GraphNode, str] = Field(default=15, description='noise level, activated for Grayscale Image Denoising and Color Image Denoising. Leave it as default or arbitrary if other tasks are selected')
    task_type: Task_type | GraphNode | tuple[GraphNode, str] = Field(default='Real-World Image Super-Resolution-Large', description='Choose a task')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.SwinIR"


from nodetool.nodes.replicate.image.upscale import Upscaler
from nodetool.nodes.replicate.image.upscale import Mode_type
from nodetool.nodes.replicate.image.upscale import Scheduler
from nodetool.nodes.replicate.image.upscale import Sampler_name
from nodetool.nodes.replicate.image.upscale import Seam_fix_mode

class UltimateSDUpscale(GraphNode):
    cfg: float | GraphNode | tuple[GraphNode, str] = Field(default=8, description='CFG')
    seed: int | None | GraphNode | tuple[GraphNode, str] = Field(default=None, description='Sampling seed, leave Empty for Random')
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    steps: int | GraphNode | tuple[GraphNode, str] = Field(default=20, description='Steps')
    denoise: float | GraphNode | tuple[GraphNode, str] = Field(default=0.2, description='Denoise')
    upscaler: Upscaler | GraphNode | tuple[GraphNode, str] = Field(default='4x-UltraSharp', description='Upscaler')
    mask_blur: int | GraphNode | tuple[GraphNode, str] = Field(default=8, description='Mask Blur')
    mode_type: Mode_type | GraphNode | tuple[GraphNode, str] = Field(default='Linear', description='Mode Type')
    scheduler: Scheduler | GraphNode | tuple[GraphNode, str] = Field(default='normal', description='Scheduler')
    tile_width: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Tile Width')
    upscale_by: float | GraphNode | tuple[GraphNode, str] = Field(default=2, description='Upscale By')
    tile_height: int | GraphNode | tuple[GraphNode, str] = Field(default=512, description='Tile Height')
    sampler_name: Sampler_name | GraphNode | tuple[GraphNode, str] = Field(default='euler', description='Sampler')
    tile_padding: int | GraphNode | tuple[GraphNode, str] = Field(default=32, description='Tile Padding')
    seam_fix_mode: Seam_fix_mode | GraphNode | tuple[GraphNode, str] = Field(default='None', description='Seam Fix Mode')
    seam_fix_width: int | GraphNode | tuple[GraphNode, str] = Field(default=64, description='Seam Fix Width')
    negative_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='', description='Negative Prompt')
    positive_prompt: str | GraphNode | tuple[GraphNode, str] = Field(default='Hey! Have a nice day :D', description='Positive Prompt')
    seam_fix_denoise: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='Seam Fix Denoise')
    seam_fix_padding: int | GraphNode | tuple[GraphNode, str] = Field(default=16, description='Seam Fix Padding')
    seam_fix_mask_blur: int | GraphNode | tuple[GraphNode, str] = Field(default=8, description='Seam Fix Mask Blur')
    controlnet_strength: float | GraphNode | tuple[GraphNode, str] = Field(default=1, description='ControlNet Strength')
    force_uniform_tiles: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Force Uniform Tiles')
    use_controlnet_tile: bool | GraphNode | tuple[GraphNode, str] = Field(default=True, description='Use ControlNet Tile')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.UltimateSDUpscale"


from nodetool.nodes.replicate.image.upscale import Scale

class ruDallE_SR(GraphNode):
    image: ImageRef | GraphNode | tuple[GraphNode, str] = Field(default=ImageRef(type='image', uri='', asset_id=None), description='Input image')
    scale: Scale | GraphNode | tuple[GraphNode, str] = Field(default=4, description='Choose up-scaling factor')
    @classmethod
    def get_node_type(cls): return "replicate.image.upscale.ruDallE_SR"


